""" general ast util functions and classes """
from dataclasses import dataclass
from functools import partial
from itertools import chain, combinations, product
from typing import Any, Callable, Collection, Iterable, Iterator, NamedTuple, Optional, Sequence, TypeVar

import networkx as nx
from clingo import SymbolType
from clingo.ast import (
    AST,
    ASTType,
    BinaryOperator,
    ComparisonOperator,
    Guard,
    Location,
    Position,
    Sign,
    Transformer,
    UnaryOperator,
)

LOC = Location(Position("<string>", 1, 1), Position("<string>", 1, 1))
SIGNS = frozenset({Sign.NoSign, Sign.Negation, Sign.DoubleNegation})

# pylint: disable=too-many-lines


@dataclass(frozen=True, order=True, eq=True)
class Predicate:
    """an (immutable) predicate consisting of a name and arity"""

    name: str
    arity: int

    def __str__(self) -> str:
        """Returns a string with the name/arity notation."""
        return f"{self.name}/{str(self.arity)}"


SignedPredicate = NamedTuple("SignedPredicate", [("sign", Sign), ("pred", Predicate)])

SignSetType = frozenset[Sign] | set[Sign]


@dataclass(frozen=True, order=True, eq=True)
class AnnotatedPredicate:
    """predicate with a list of numbers of position
    ProjectedPred(foo/4), [0,3] represents something like foo(X,_,_,Y)
    also contains the line number of creation"""

    pred: Predicate
    annotated_positions: tuple[int, ...]


def negate_comparison(cmp: ComparisonOperator) -> ComparisonOperator:
    """reverse clingo.ast.ComparisonOperator"""
    return {
        ComparisonOperator.Equal: ComparisonOperator.NotEqual,
        ComparisonOperator.NotEqual: ComparisonOperator.Equal,
        ComparisonOperator.GreaterEqual: ComparisonOperator.LessThan,
        ComparisonOperator.LessEqual: ComparisonOperator.GreaterThan,
        ComparisonOperator.GreaterThan: ComparisonOperator.LessEqual,
        ComparisonOperator.LessThan: ComparisonOperator.GreaterEqual,
    }[cmp]


def rhs2lhs_comparison(cmp: ComparisonOperator) -> ComparisonOperator:
    """move clingo.ast.ComparisonOperator from rhs to lhs"""
    return {
        ComparisonOperator.Equal: ComparisonOperator.Equal,
        ComparisonOperator.NotEqual: ComparisonOperator.NotEqual,
        ComparisonOperator.GreaterEqual: ComparisonOperator.LessEqual,
        ComparisonOperator.LessEqual: ComparisonOperator.GreaterEqual,
        ComparisonOperator.GreaterThan: ComparisonOperator.LessThan,
        ComparisonOperator.LessThan: ComparisonOperator.GreaterThan,
    }[cmp]


def compare(lhs: int, cmp: ComparisonOperator, rhs: int) -> bool:
    """compare two integers using the AST comparison operator"""
    if cmp == ComparisonOperator.Equal:
        return lhs == rhs
    if cmp == ComparisonOperator.NotEqual:
        return lhs != rhs
    if cmp == ComparisonOperator.GreaterEqual:
        return lhs >= rhs
    if cmp == ComparisonOperator.LessEqual:
        return lhs <= rhs
    if cmp == ComparisonOperator.GreaterThan:
        return lhs > rhs
    if cmp == ComparisonOperator.LessThan:
        return lhs < rhs
    assert False, "unknown Comparison Operator used"


def negate_agg(agg: AST) -> AST:
    """negate the guards of an aggregate in the body"""
    assert agg.ast_type in (ASTType.BodyAggregate, ASTType.Aggregate)
    agg = agg.update()
    if agg.left_guard:
        agg = agg.update(left_guard=agg.left_guard.update(comparison=negate_comparison(agg.left_guard.comparison)))
    if agg.right_guard:  # nocoverage not produced by sympy
        agg = agg.update(right_guard=agg.right_guard.update(comparison=negate_comparison(agg.right_guard.comparison)))
    return agg


# TODO: refactor, not make it specific for equal variables, do not have names and guards as different bounds etc...
# also extend to HeadAggBounds as used in (or done manually) in sum_aggregates.py
@dataclass
class AggAnalytics:
    """class to analyze a body aggregate and capture its bounds"""

    def __init__(self, node: AST):
        assert node.ast_type in (ASTType.BodyAggregate, ASTType.HeadAggregate, ASTType.Aggregate)
        self.equal_variable_bound: list[str] = []  # list of all equal variables as variable names
        self.bounds: list[AST] = []  # all non equal variable bounds as right guards

        if node.left_guard and node.left_guard.ast_type == ASTType.Guard:
            guard = node.left_guard
            if guard.comparison == ComparisonOperator.Equal and guard.term.ast_type == ASTType.Variable:
                self.equal_variable_bound.append(guard.term.name)
            else:
                self.bounds.append(Guard(rhs2lhs_comparison(guard.comparison), guard.term))

        if node.right_guard and node.right_guard.ast_type == ASTType.Guard:
            guard = node.right_guard
            if guard.comparison == ComparisonOperator.Equal and guard.term.ast_type == ASTType.Variable:
                self.equal_variable_bound.append(guard.term.name)
            else:
                self.bounds.append(guard)

    def guaranteed_leq(self, number: int) -> bool:
        """return True if it can be identified that the upper bound of the aggregate is leq number (or below)"""
        for bound in self.bounds:
            if bound.term.ast_type == ASTType.SymbolicTerm and bound.term.symbol.type == SymbolType.Number:
                if bound.comparison in (ComparisonOperator.LessEqual, ComparisonOperator.Equal):
                    return int(bound.term.symbol.number) <= number
                if bound.comparison == ComparisonOperator.LessThan:
                    return int(bound.term.symbol.number) - 1 <= number
        return False

    def guaranteed_geq(self, number: int) -> bool:
        """return True if it can be identified that the lower bound of the aggregate is geq number (or above)"""
        for bound in self.bounds:
            if bound.term.ast_type == ASTType.SymbolicTerm and bound.term.symbol.type == SymbolType.Number:
                if bound.comparison in (ComparisonOperator.GreaterEqual, ComparisonOperator.Equal):
                    return int(bound.term.symbol.number) >= number
                if bound.comparison == ComparisonOperator.GreaterThan:
                    return int(bound.term.symbol.number) + 1 >= number
        return False


class GeneralVisitor(Transformer):
    """helper class to visit specific asts"""

    def __init__(self, ast_name: str):
        self.collection: list[AST] = []
        setattr(self, "visit_" + ast_name, self.visitnode)

    def visitnode(self, node: AST) -> AST:
        """generic visit function"""
        self.collection.append(node)
        return node


def collect_ast(stm: AST, ast_name: str) -> list[AST]:
    """returns the list of asts of the given ast_name inside the stm"""
    visitor = GeneralVisitor(ast_name)
    visitor.visit(stm)
    return visitor.collection


class GeneralTransformer(Transformer):
    """helper class to transform specific asts"""

    def __init__(self, ast_name: str, function: Callable[[AST], AST]):
        self.function = function
        setattr(self, "visit_" + ast_name, self.visitnode)

    def visitnode(self, node: AST) -> AST:
        """generic visit function"""
        return self.function(node)


def transform_ast(stm: AST, ast_name: str, function: Callable[[AST], AST]) -> AST:
    """transforms all ast with name ast_name in stm by calling function on them"""
    transformer = GeneralTransformer(ast_name, function)
    return transformer(stm)


def _potentially_unifying(lhs: AST, rhs: AST) -> bool:
    if (lhs == rhs) or (ASTType.Variable in (lhs.ast_type, rhs.ast_type)):
        return True
    nfunc = {
        ASTType.SymbolicTerm,
        ASTType.UnaryOperation,
        ASTType.BinaryOperation,
        ASTType.Interval,
    }
    if rhs.ast_type in nfunc:
        rhs, lhs = lhs, rhs

    if rhs.ast_type == ASTType.Function and lhs.ast_type in nfunc:
        return False

    if lhs.ast_type == ASTType.SymbolicTerm and rhs.ast_type == ASTType.SymbolicTerm:
        return bool(lhs == rhs)

    if lhs.ast_type == ASTType.UnaryOperation and (
        rhs.ast_type == ASTType.UnaryOperation and lhs.operator_type == rhs.operator_type
    ):
        return _potentially_unifying(lhs.argument, rhs.argument)

    if lhs.ast_type == ASTType.Function and rhs.ast_type == ASTType.Function:
        return (
            lhs.name == rhs.name
            and len(lhs.arguments) == len(rhs.arguments)
            and all(
                map(
                    lambda x: _potentially_unifying(x[0], x[1]),
                    zip(lhs.arguments, rhs.arguments),
                )
            )
        )

    return True


def potentially_unifying_sequence(lhs: Sequence[AST], rhs: Sequence[AST]) -> bool:
    """
    returns True if both sequences are potentially unifying
    see @potentially_unifying
    """
    if len(lhs) != len(rhs):
        return False
    return all(
        map(
            lambda x: potentially_unifying(*x),
            zip(lhs, rhs),
        )
    )


def potentially_unifying(lhs: AST, rhs: AST) -> bool:
    """returns True if both terms could potentially be equal after variable substitution
    Conservative, may falsely return True in cases it does not know
    Also does not consider Variable Names, so f(X) could unify with X
    """
    terms = {
        ASTType.SymbolicTerm,
        ASTType.Variable,
        ASTType.UnaryOperation,
        ASTType.BinaryOperation,
        ASTType.Interval,
        ASTType.Function,
        ASTType.Pool,
    }
    assert lhs.ast_type in terms
    assert rhs.ast_type in terms

    if any(
        map(
            lambda x: _potentially_unifying(x[0], x[1]),
            product(lhs.unpool(), rhs.unpool()),
        )
    ):
        return True
    return False


def body_predicates(rule: AST, signs: SignSetType) -> Iterator[SignedPredicate]:
    """
    yields all predicates used in the rule body as (name, arity) that have a sign in the set signs
    """
    if rule.ast_type == ASTType.Rule:
        for blit in rule.body:
            yield from literal_predicate(blit, signs)
            yield from conditional_literal_predicate(blit, signs)


def minimize_predicates(stm: AST, signs: SignSetType) -> Iterator[SignedPredicate]:
    """
    yields all predicates used in the minimize statement as (name, arity) that have a sign in the set signs
    """
    if stm.ast_type == ASTType.Minimize:
        for blit in stm.body:
            if blit.ast_type == ASTType.Literal:
                yield from literal_predicate(blit, signs)
            yield from conditional_literal_predicate(blit, signs)


def literal_predicate(lit: AST, signs: SignSetType) -> Iterator[SignedPredicate]:
    """converts ast Literal into (sign, name, arity) if sign is in signs"""
    if lit.ast_type == ASTType.Literal:
        atom = lit.atom
        if lit.sign in signs and atom.ast_type == ASTType.SymbolicAtom:
            if atom.symbol.ast_type == ASTType.Function:
                yield SignedPredicate(lit.sign, Predicate(atom.symbol.name, len(atom.symbol.arguments)))
        yield from aggregate_predicate(atom, signs)
        yield from headorbody_aggregate_predicate(atom, signs)
        yield from conditional_literal_predicate(atom, signs)


def conditional_literal_predicate(condlit: AST, signs: SignSetType) -> Iterator[SignedPredicate]:
    """
    yields all predicates used in the conditional literal as (name, arity) that have a sign in the set signs
    """
    if condlit.ast_type != ASTType.ConditionalLiteral:
        return
    lit = condlit.literal
    yield from literal_predicate(lit, signs)
    for cond in condlit.condition:
        yield from literal_predicate(cond, signs)


def headorbody_aggregate_predicate(agg: AST, signs: SignSetType) -> Iterator[SignedPredicate]:
    """
    yields all predicates used in the head or body agregate as (name, arity) that have a sign in the set signs
    """
    if agg.ast_type in {ASTType.BodyAggregate, ASTType.HeadAggregate}:
        for elem in agg.elements:
            if elem.ast_type == ASTType.HeadAggregateElement:
                yield from conditional_literal_predicate(elem.condition, signs)
            elif elem.ast_type == ASTType.BodyAggregateElement:
                for cond in elem.condition:
                    yield from literal_predicate(cond, signs)


def aggregate_predicate(agg: AST, signs: SignSetType) -> Iterator[SignedPredicate]:
    """
    yields all predicates used in the aggregate as (name, arity) that have a sign in the set signs
    """
    if agg.ast_type == ASTType.Aggregate:
        for elem in agg.elements:
            yield from conditional_literal_predicate(elem, signs)
            for cond in elem.condition:
                # aggregate in body seems to have Literals as condition
                yield from literal_predicate(cond, signs)


def disjunction_predicate(head: AST, signs: SignSetType) -> Iterator[SignedPredicate]:
    """
    yields all predicates used in the disjunction head as (name, arity) that have a sign in the set signs
    """
    if head.ast_type == ASTType.Disjunction:
        for lit in head.elements:
            yield from conditional_literal_predicate(lit, signs)


def head_predicates(rule: AST, signs: SignSetType) -> Iterator[SignedPredicate]:
    """
    yields all predicates used in the rule head as (name, arity) that have a sign in the set signs
    """
    if rule.ast_type == ASTType.Rule:
        head = rule.head
        yield from literal_predicate(head, signs)
        yield from aggregate_predicate(head, signs)
        yield from headorbody_aggregate_predicate(head, signs)
        yield from disjunction_predicate(head, signs)


def __get_preds_from_literal_in_conditional(condition: AST, signs: SignSetType) -> Iterator[SignedPredicate]:
    assert condition.ast_type == ASTType.ConditionalLiteral
    yield from literal_predicate(condition.literal, signs)


def headderivable_predicates(rule: AST) -> Iterator[SignedPredicate]:
    """
    yields all predicates used in the rule head that are derivable
    """

    positive = {Sign.NoSign}
    if rule.ast_type == ASTType.Rule:
        head = rule.head
        yield from literal_predicate(head, positive)
        if head.ast_type == ASTType.Aggregate:
            for elem in head.elements:
                yield from __get_preds_from_literal_in_conditional(elem, positive)
        if head.ast_type == ASTType.HeadAggregate:
            for elem in head.elements:
                if elem.ast_type == ASTType.HeadAggregateElement:
                    yield from __get_preds_from_literal_in_conditional(elem.condition, positive)
        if head.ast_type == ASTType.Disjunction:
            for lit in head.elements:
                yield from __get_preds_from_literal_in_conditional(lit, positive)


def predicates(ast: AST, signs: SignSetType = SIGNS) -> Iterator[SignedPredicate]:
    """
    yields all predicates in ast that have a sign in the set signs
    """
    yield from head_predicates(ast, signs)
    yield from literal_predicate(ast, signs)
    yield from aggregate_predicate(ast, signs)
    yield from headorbody_aggregate_predicate(ast, signs)
    yield from disjunction_predicate(ast, signs)
    yield from conditional_literal_predicate(ast, signs)
    yield from body_predicates(ast, signs)
    yield from minimize_predicates(ast, signs)


def conditions_of_body_agg(agg: AST) -> list[AST]:
    """return all conditions inside an body aggregate in a list"""
    ret: list[AST] = []
    if agg.ast_type in (ASTType.BodyAggregate, ASTType.Aggregate):
        for elem in agg.elements:
            if elem.condition:
                ret.extend(elem.condition)
    return ret


def has_interval(ast: AST) -> bool:
    """true if ast contains an interval"""
    return bool(collect_ast(ast, "Interval"))


def has_unsafe_operation(ast: AST) -> bool:
    """true if ast contains a possibly unsafe operation"""
    if collect_ast(ast, "Interval"):
        return True

    if any(map(lambda x: x.operator_type == UnaryOperator.Absolute, collect_ast(ast, "UnaryOperation"))):
        return True

    invalid = (
        BinaryOperator.XOr,
        BinaryOperator.Power,
        BinaryOperator.Modulo,
        BinaryOperator.Division,
        BinaryOperator.Multiplication,
    )
    return any(map(lambda x: x.operator_type in invalid, collect_ast(ast, "BinaryOperation")))


def _collect_binding_information_simple_literal(
    lit: AST, in_bound_vars: set[AST], in_unbound_vars: set[AST]
) -> tuple[set[AST], set[AST]]:
    bound_variables: set[AST] = set(in_bound_vars)
    unbound_variables: set[AST] = set(in_unbound_vars)
    assert lit.ast_type == ASTType.Literal
    if lit.atom.ast_type == ASTType.SymbolicAtom:
        # simple operations (no absolute with more than 1 variable) can bind exactly one variable
        # if all other variables are bound
        # This is probably not exactly the thing that gringo does
        if lit.sign == Sign.NoSign and lit.atom.symbol.ast_type == ASTType.Function:
            for arg in lit.atom.symbol.arguments:
                variables = collect_ast(arg, "Variable")
                if (
                    len(variables) == 1
                    and not has_unsafe_operation(arg)
                    or len(collect_ast(arg, "BinaryOperation")) + len(collect_ast(arg, "UnaryOperation")) == 0
                ):
                    bound_variables.update(variables)
                else:
                    unbound_variables.update(variables)
        else:
            unbound_variables.update(collect_ast(lit, "Variable"))
    elif lit.atom.ast_type == ASTType.Comparison:
        bound, unbound = _collect_binding_information_from_comparison(lit, bound_variables)
        bound_variables.update(bound)
        unbound_variables.update(unbound)
    return bound_variables, unbound_variables


def _collect_binding_information_conditions(
    conditions: Iterable[AST], already_bound: set[AST]
) -> tuple[set[AST], set[AST]]:
    """given a list of Literal inside a condition
    returns a set of Variables that it binds
    returns a set of Variables that it needs to be bounded
    additional input: a set of already bound variables"""
    bound_variables: set[AST] = set(already_bound)
    unbound_variables: set[AST] = set()
    size = -1
    while len(bound_variables) != size:
        size = len(bound_variables)
        for condition in conditions:
            bound, unbound = _collect_binding_information_simple_literal(condition, bound_variables, unbound_variables)
            bound_variables.update(bound)
            unbound_variables.update(unbound)
    unbound_variables -= bound_variables
    return bound_variables, unbound_variables


def _collect_binding_information_from_equal(
    lhs: AST, rhs: AST, input_bound_variables: set[AST]
) -> tuple[set[AST], set[AST]]:
    bound_variables: set[AST] = input_bound_variables
    unbound_variables: set[AST] = set(collect_ast(lhs, "Variable")) | set(collect_ast(rhs, "Variable"))
    if all(i.ast_type == ASTType.Function and i.name == "" and i.arguments for i in (lhs, rhs)) and len(
        lhs.arguments
    ) == len(rhs.arguments):
        for left, right in zip(lhs.arguments, rhs.arguments):
            bound, unbound = _collect_binding_information_from_equal(left, right, bound_variables)
            bound_variables.update(bound)
            unbound_variables.update(unbound)
    else:
        lhs_vars = set(collect_ast(lhs, "Variable"))
        rhs_vars = set(collect_ast(rhs, "Variable"))
        if len(lhs_vars) == 1 and not has_unsafe_operation(lhs) and rhs_vars <= bound_variables:
            bound_variables.update(lhs_vars)
        if len(rhs_vars) == 1 and not has_unsafe_operation(rhs) and lhs_vars <= bound_variables:
            bound_variables.update(rhs_vars)
    return bound_variables, unbound_variables - bound_variables


def comparison2comparisonlist(comparison: AST) -> list[tuple[AST, ComparisonOperator, AST]]:
    """convert the nested AST Comparison structure to a plain list of (term op term)"""
    assert comparison.ast_type == ASTType.Comparison
    ret: list[tuple[AST, ComparisonOperator, AST]] = []
    lhs = comparison.term
    for guard in comparison.guards:
        operator = guard.comparison
        rhs = guard.term
        ret.append((lhs, operator, rhs))
        lhs = rhs
    return ret


def _collect_binding_information_from_comparison(
    comparison: AST, input_bound_variables: set[AST]
) -> tuple[set[AST], set[AST]]:
    assert comparison.ast_type == ASTType.Literal
    assert comparison.atom.ast_type == ASTType.Comparison
    if comparison.sign != Sign.NoSign:
        return (set(), set(collect_ast(comparison.atom, "Variable")))
    bound_variables: set[AST] = set(input_bound_variables)
    unbound_variables: set[AST] = set(collect_ast(comparison, "Variable"))
    for lhs, operator, rhs in comparison2comparisonlist(comparison.atom):
        if operator == ComparisonOperator.Equal:
            bound, unbound = _collect_binding_information_from_equal(lhs, rhs, bound_variables)
            bound_variables.update(bound)
            unbound_variables.update(unbound)
    return bound_variables, unbound_variables - bound_variables


def _collect_binding_information_from_comparisons(
    stmlist: Iterable[AST], input_bound_variables: set[AST]
) -> tuple[set[AST], set[AST]]:
    """given a list of body literal and already bound variables
    returns a set of Variables that it binds by comparison ASTs
    returns a set of Variables that it needs to be bounded by comparison ASTs"""
    # needs to run until a fixpoint is found
    bound_variables: set[AST] = input_bound_variables
    unbound_variables: set[AST] = set()
    while True:
        orig = bound_variables.copy()
        for stm in stmlist:
            if stm.ast_type == ASTType.Literal and stm.atom.ast_type == ASTType.Comparison:
                bound, unbound = _collect_binding_information_from_comparison(stm, bound_variables)
                bound_variables.update(bound)
                unbound_variables.update(unbound)

        if orig == bound_variables:
            break

    return bound_variables, unbound_variables


def collect_binding_information_head(head: AST, body: list[AST]) -> tuple[set[AST], set[AST]]:
    """given a head
    returns a set of Variables that it needs to be bounded
    returns a set of Variables that does not need to be bounded"""
    # pylint: disable=too-many-nested-blocks
    bound_in_body = collect_binding_information_body(body)[0]
    need_bound_variables: set[AST] = set()
    no_bound_needed: set[AST] = set()

    if head.ast_type == ASTType.Literal:
        need_bound_variables.update(collect_ast(head, "Variable"))
    if head.ast_type in (ASTType.HeadAggregate, ASTType.Aggregate):
        if head.left_guard is not None:
            need_bound_variables.update(collect_ast(head.left_guard, "Variable"))
        if head.right_guard is not None:
            need_bound_variables.update(collect_ast(head.right_guard, "Variable"))
        for element in head.elements:
            term_vars = set()
            if head.ast_type == ASTType.HeadAggregate:
                term_vars = set().union(*map(partial(collect_ast, ast_name="Variable"), element.terms))
                term_vars.update(collect_ast(element.condition.literal, "Variable"))
                bound, unbound = _collect_binding_information_conditions(element.condition.condition, bound_in_body)
                term_vars -= bound
                need_bound_variables.update(term_vars)
                need_bound_variables.update(unbound)
            if head.ast_type == ASTType.Aggregate:
                bound, unbound = _collect_binding_information_conditions(element.condition, bound_in_body)
                need_bound_l = set(collect_ast(element.literal, "Variable"))
                need_bound_variables.update(need_bound_l - bound)
                need_bound_variables.update(unbound)
            no_bound_needed.update(bound)

    elif head.ast_type == ASTType.Disjunction:
        for element in head.elements:
            bound, unbound = _collect_binding_information_conditions(element.condition, bound_in_body)
            need_bound_l = set(collect_ast(element.literal, "Variable"))
            need_bound_variables.update(need_bound_l - bound)
            no_bound_needed.update(bound)
    need_bound_variables = set(filter(lambda var: var.name != "_", need_bound_variables))
    no_bound_needed = set(filter(lambda var: var.name != "_", no_bound_needed))
    need_bound_variables -= bound_in_body
    no_bound_needed |= bound_in_body
    return need_bound_variables, no_bound_needed


def collect_binding_information_body(
    stmlist: Iterable[AST], prebound: Optional[set[AST]] = None
) -> tuple[set[AST], set[AST]]:
    """given a list of body literal
    returns a set of Variables that it binds
    returns a set of Variables that it needs to be bounded"""
    # pylint: disable=too-many-nested-blocks, too-many-branches
    bound_variables: set[AST] = set()
    if prebound is not None:
        bound_variables.update(prebound)
    unbound_variables: set[AST] = set()
    ### need to do a fixpoint computation
    size_before = -1
    while len(bound_variables) > size_before:
        for stm in stmlist:
            if stm.ast_type == ASTType.Literal:
                bound, unbound = _collect_binding_information_simple_literal(stm, bound_variables, unbound_variables)
                bound_variables.update(bound)
                unbound_variables.update(unbound)
                if stm.atom.ast_type in (ASTType.BodyAggregate, ASTType.Aggregate):
                    if stm.atom.left_guard is not None:
                        if stm.sign == Sign.NoSign and stm.atom.left_guard.comparison == ComparisonOperator.Equal:
                            bound_variables.update(collect_ast(stm.atom.left_guard, "Variable"))
                        else:
                            unbound_variables.update(collect_ast(stm.atom.left_guard, "Variable"))
                    if stm.atom.right_guard is not None:
                        if stm.sign == Sign.NoSign and stm.atom.right_guard.comparison == ComparisonOperator.Equal:
                            bound_variables.update(collect_ast(stm.atom.right_guard, "Variable"))
                        else:
                            unbound_variables.update(collect_ast(stm.atom.right_guard, "Variable"))
                    for element in stm.atom.elements:
                        term_vars = set()
                        if stm.atom.ast_type == ASTType.BodyAggregate:
                            term_vars = set().union(*map(partial(collect_ast, ast_name="Variable"), element.terms))
                            bound, unbound = _collect_binding_information_conditions(element.condition, bound_variables)
                        assert stm.atom.ast_type != ASTType.Aggregate

                        term_vars -= bound
                        term_vars -= bound_variables
                        unbound_variables.update(term_vars)
                        unbound -= bound_variables
                        unbound_variables.update(unbound)
            elif stm.ast_type == ASTType.ConditionalLiteral:
                term_vars = set(collect_ast(stm.literal, ast_name="Variable"))
                bound, unbound = _collect_binding_information_conditions(stm.condition, bound_variables)
                unbound_variables.update(unbound)
                unbound_variables.update(term_vars - bound)
        unbound_variables -= bound_variables
        bound, unbound = _collect_binding_information_from_comparisons(stmlist, bound_variables)
        bound_variables.update(bound)
        unbound_variables.update(unbound)
        unbound_variables -= bound_variables
        size_before = len(bound_variables)
    bound_variables = set(filter(lambda var: var.name != "_", bound_variables))
    unbound_variables = set(filter(lambda var: var.name != "_", unbound_variables))
    return bound_variables, unbound_variables


def collect_bound_variables(stmlist: Iterable[AST]) -> set[AST]:
    """return a set of all ast of type Variable that are in a positive symbolic literal or in a eq relation"""
    return collect_binding_information_body(stmlist)[0]


def loc2str(loc: Location) -> str:
    """format location to be a nice looking string"""
    return f"{loc.begin.filename}:{loc.begin.line}:{loc.begin.column}"


def global_vars_inside_body(lits: list[AST]) -> set[AST]:
    """compute all Variables that are visible within the same outer scope"""
    return set.union(*collect_binding_information_body(lits))


def global_vars_inside_head(head: AST) -> set[AST]:
    """compute all Variables that are visible within the same outer scope"""
    return set.union(*collect_binding_information_head(head, []))


def _get_simple_equalities(lits: list[AST]) -> list[AST]:
    """only return var1 = var2 or not var1 != var2 equalities"""
    ret: list[AST] = []
    for lit in lits:
        if (
            lit.ast_type == ASTType.Literal
            and lit.atom.ast_type == ASTType.Comparison
            and lit.atom.term.ast_type == ASTType.Variable
            and lit.atom.guards[0].term.ast_type == ASTType.Variable
        ):
            if (lit.sign == Sign.NoSign and lit.atom.guards[0].comparison == ComparisonOperator.Equal) or (
                lit.sign == Sign.Negation and lit.atom.guards[0].comparison == ComparisonOperator.NotEqual
            ):
                ret.append(lit)
    return ret


def _replace(uniques: list[tuple[set[AST], AST]], var: AST) -> AST:
    for cc, single in uniques:
        if var in cc:
            return single
    return var


def replace_simple_assignments_aggregate(lit: AST) -> AST:
    """replace variable equalities with their inlined versions
    e.g. foo(X), bar(Y), X=Y becomes foo(X), bar(X) inside an aggregate"""
    assert lit.atom.ast_type == ASTType.BodyAggregate
    new_elements: list[AST] = []
    for elem in lit.atom.elements:
        eqs = _get_simple_equalities(elem.condition)
        graph = nx.Graph()
        for eq in eqs:
            graph.add_edge(eq.atom.term, eq.atom.guards[0].term)
        uniques: list[tuple[set[AST], AST]] = []
        for cc in nx.connected_components(graph):
            uniques.append((cc, sorted(cc)[0]))
        new_condition = [c for c in elem.condition if c not in eqs]
        new_elem = elem.update(condition=new_condition)
        new_elements.append(transform_ast(new_elem, "Variable", partial(_replace, uniques)))
    new_atom = lit.atom.update(elements=new_elements)
    return lit.update(atom=new_atom)


def replace_simple_assignments(stm: AST) -> AST:
    """replace variable equalities with their inlined versions
    e.g. foo(X), bar(Y), X=Y becomes foo(X), bar(X)"""
    if stm.ast_type not in (ASTType.Rule, ASTType.Minimize):
        return stm
    literals: Iterable[AST] = stm.body
    new_body: list[AST] = []
    new_heads: list[AST]
    if stm.ast_type == ASTType.Rule:
        new_heads = [stm.head]
    else:
        new_heads = [stm.weight, stm.priority, *stm.terms]
    # normalize comparison operators
    new_body = list(literals)
    graph = nx.Graph()
    aux_body: list[AST] = []
    eqs = _get_simple_equalities(new_body)
    for lit in new_body:
        if lit in eqs:
            continue
        if lit.ast_type == ASTType.Literal and lit.atom.ast_type == ASTType.BodyAggregate:
            aux_body.append(replace_simple_assignments_aggregate(lit))
        else:
            aux_body.append(lit)

    for eq in eqs:
        graph.add_edge(eq.atom.term, eq.atom.guards[0].term)
    uniques: list[tuple[set[AST], AST]] = []
    for cc in nx.connected_components(graph):
        uniques.append((cc, sorted(cc)[0]))

    p = partial(transform_ast, ast_name="Variable", function=partial(_replace, uniques))
    aux_body = list(map(p, aux_body))
    new_heads = list(map(p, new_heads))

    # for lit in aux_body:
    #     new_body.append(transform_ast(lit, "Variable", partial(_replace, uniques)))

    if stm.ast_type == ASTType.Rule:
        return stm.update(head=new_heads[0], body=aux_body)
    return stm.update(weight=new_heads[0], priority=new_heads[1], terms=new_heads[2:], body=aux_body)


def replace_assignments(stm: AST) -> AST:
    """replace equalities with their inlined versions
    e.g. foo(X), bar(Y), X=Y becomes foo(X), bar(X)"""
    if stm.ast_type not in (ASTType.Rule, ASTType.Minimize):
        return stm
    literals: Iterable[AST] = stm.body
    new_body: list[AST] = []
    new_heads: list[AST]
    if stm.ast_type == ASTType.Rule:
        new_heads = [stm.head]
    else:
        new_heads = [stm.weight, stm.priority, *stm.terms]

    new_body = list(literals)
    removal: list[int] = []
    for index, lit in enumerate(new_body):
        if (
            lit.ast_type == ASTType.Literal
            and lit.atom.ast_type == ASTType.Comparison
            and lit.atom.term.ast_type == ASTType.Variable
            and not has_interval(lit.atom.guards[0].term)
        ):
            if (lit.sign == Sign.NoSign and lit.atom.guards[0].comparison == ComparisonOperator.Equal) or (
                lit.sign == Sign.Negation and lit.atom.guards[0].comparison == ComparisonOperator.NotEqual
            ):
                for other, elem in enumerate(new_body):
                    if other == index:
                        continue
                    new_body[other] = transform_ast(
                        elem, "Variable", partial(replace_var_name, lit.atom.term, lit.atom.guards[0].term)
                    )
                removal.append(index)
                for i, head in enumerate(new_heads):
                    new_heads[i] = transform_ast(
                        head, "Variable", partial(replace_var_name, lit.atom.term, lit.atom.guards[0].term)
                    )
                continue
    for index in reversed(removal):
        new_body.pop(index)
    if stm.ast_type == ASTType.Rule:
        return stm.update(head=new_heads[0], body=new_body)
    return stm.update(weight=new_heads[0], priority=new_heads[1], terms=new_heads[2:], body=new_body)


def replace_var_name(orig: AST, replace: AST, var: AST) -> AST:
    """replace orig variable with replace, if var == orig"""
    assert orig.ast_type == ASTType.Variable
    if var == orig:
        return replace
    return var


class TranslationMap:
    """translates an old predicate to a new one"""

    def __init__(self, oldpred: Predicate, newpred: Predicate, mapping: Iterable[int | None]):
        self.oldpred = oldpred
        self.newpred = newpred
        # simple ordered list of indices or none, to map f(A1,A2,A4) to b(A1,A4,A3,A2)
        # have mapping [0,3,1], reverse mapping would be [0,2,None,1]
        self.mapping = mapping

    def translate_parameters(self, arguments: list[Any]) -> list[AST | None]:
        """given the mapping, return the mapped order of the argument terms"""
        ret: list[AST | None] = []
        for oldidx, index in enumerate(self.mapping):
            if index is None:
                continue
            assert len(arguments) > index
            if index >= len(ret):
                ret.extend([None] * (index + 1 - len(ret)))
            ret[index] = arguments[oldidx]
        return ret


def is_predicate(lit: AST) -> bool:
    """true if lit is a literal with a named predicate"""
    if lit.ast_type == ASTType.Literal:
        atom = lit.atom
        return bool(atom.ast_type == ASTType.SymbolicAtom and atom.symbol.ast_type == ASTType.Function)
    return False


def is_comparison(lit: AST) -> bool:
    """true if lit is a literal with a named predicate"""
    return lit.ast_type == ASTType.Literal and lit.atom.ast_type == ASTType.Comparison


def is_body_aggregate(lit: AST) -> bool:
    """true if lit is a literal with a named predicate"""
    return lit.ast_type == ASTType.Literal and lit.atom.ast_type == ASTType.BodyAggregate


def is_conditional(lit: AST) -> bool:
    """true if lit is a literal with a named predicate"""
    return lit.ast_type == ASTType.ConditionalLiteral


T = TypeVar("T")


def largest_subset(input_list: Collection[T]) -> list[Collection[T]]:
    """return all subsets of the input list in decreasing size"""
    return list(reversed(list(chain.from_iterable(combinations(input_list, r) for r in range(len(input_list) + 1)))))



""" test projecting predicates """

import pytest
from clingo.ast import AST, parse_string

from ngo.normalize import postprocess, preprocess
from ngo.projection import ProjectionTranslator
from ngo.utils.ast import Predicate


@pytest.mark.parametrize(
    "lhs, input_predicates, rhs",
    (
        (  # nothing to project
            """
{ a((1..3)) }.
foo(X) :- inline(X,Y), not bar(X,Y).
#show bar : foo(X).
            """,
            [],
            """#program base.
{ a((1..3)) }.
foo(X) :- inline(X,Y); not bar(X,Y).
#show bar : foo(X).""",
        ),
        (
            """
p(A,D) :- q(A,B,C), r(A,D), t(E), not s(B,E).
            """,
            [],
            """#program base.
__aux_1(A) :- q(A,B,C); t(E); not s(B,E).
p(A,D) :- r(A,D); __aux_1(A).""",
        ),
        (
            """
#false :- job(J); machine(M); 2 <= #sum { 1,0,perm(J,M,P): perm(J,M,P) }.
            """,
            [],
            """#program base.
#false :- job(J); machine(M); 2 <= #sum { 1,0,perm(J,M,P): perm(J,M,P) }.""",
        ),
        (
            """
blockedy(Z,W,T) :- at(X,Y,T); blockedx(Z,W,T); minom(Z,W,T); wallside(Z,W,(0,((Y-W)/|(Y-W)|))); not Y = W.
            """,
            [],
            """#program base.
blockedy(Z,W,T) :- at(X,Y,T); blockedx(Z,W,T); minom(Z,W,T); wallside(Z,W,(0,((Y-W)/|(Y-W)|))); not Y = W.""",
        ),
        (
            """
p(A,D) :- q(A,B,_), r(A,B), not t(B).
            """,
            [],
            """#program base.
p(A,D) :- q(A,B,_); r(A,B); not t(B).""",
        ),
        (
            """
p(A,D) :- q(A,B,C); r(A,D); t(B,D).
            """,
            [],
            """#program base.
p(A,D) :- q(A,B,C); r(A,D); t(B,D).""",
        ),
        (
            """
p(A,D) :- q(A,B,C), r(A,D), E = #sum { 1 }, not s(B,E), D = #sum { 2 }.
            """,
            [],
            """#program base.
p(A,D) :- q(A,B,C); r(A,D); E = #sum { 1 }; not s(B,E); D = #sum { 2 }.""",
        ),
    ),
)
def test_projection_translation(lhs: str, input_predicates: list[Predicate], rhs: str) -> None:
    """test inlining literals"""
    ast: list[AST] = []
    parse_string(lhs, ast.append)
    ast = preprocess(ast)
    utr = ProjectionTranslator(ast, input_predicates)
    ast = utr.execute(ast)
    ast = postprocess(ast)
    output = "\n".join(map(str, ast))
    assert output == rhs



""" test inlining predicates """

import pytest
from clingo.ast import AST, parse_string

from ngo.inline import InlineTranslator
from ngo.normalize import postprocess, preprocess
from ngo.utils.ast import Predicate


@pytest.mark.parametrize(
    "lhs, input_predicates, output_predicates, rhs",
    (
        (  # nothing to inline
            """
{ a((1..3)) }.
foo(X) :- inline(X,Y), not bar(X,Y).
#show bar : foo(X).
            """,
            [],
            [],
            """#program base.
{ a((1..3)) }.
foo(X) :- inline(X,Y); not bar(X,Y).
#show bar : foo(X).""",
        ),
        (  # no replacement in same aggregate as B is used in inline rule
            """
{ a((1..3)) }.
suminline(A,B) :- a(A,B), B = #sum {5;7}.
foo(X) :- X = #sum {F,V : suminline(V,F); A,B : test(A,B)}.
            """,
            [],
            [],
            """#program base.
{ a((1..3)) }.
suminline(A,B) :- a(A,B); B = #sum { 5; 7 }.
foo(X) :- X = #sum { F,V: suminline(V,F); A,B: test(A,B) }.""",
        ),
        (  # no replacement in agg condition without aggregates
            """
{ a((1..3),2) }.
suminline(A,B) :- a(A,B).
foo(X) :- X = #sum {F,V : suminline(V,F); A,B : test(A,B)}.
            """,
            [],
            [],
            """#program base.
{ a((1..3),2) }.
suminline(A,B) :- a(A,B).
foo(X) :- X = #sum { F,V: suminline(V,F); A,B: test(A,B) }.""",
        ),
        (  # no replacement as ambigious term (F,V) = (A,B)
            """
{ a((1..3)) }.
nosuminline(A,B) :- a(A), B = #sum { Y: person(A,Y) }.
foo(X) :- X = #sum { F,V: nosuminline(V,F); A,B: test(A,B) }.
            """,
            [],
            [],
            """#program base.
{ a((1..3)) }.
nosuminline(A,B) :- a(A); B = #sum { Y: person(A,Y) }.
foo(X) :- X = #sum { F,V: nosuminline(V,F); A,B: test(A,B) }.""",
        ),
        (  # replacement in same aggregate but have condition
            """
{ a((1..3)) }.
suminline(A,B) :- a(A) : b; B = #sum { Y: person(A,Y) }.
foo(X) :- X = #sum { F,V: suminline(V,F); A: test(A,B) }.
            """,
            [],
            [],
            """#program base.
{ a((1..3)) }.
suminline(A,B) :- a(A): b; B = #sum { Y: person(A,Y) }.
foo(X) :- X = #sum { F,V: suminline(V,F); A: test(A,B) }.""",
        ),
        (  # replacement in same aggregate, agg result and weight have to match, could be improved
            """
{ a((1..3)) }.
suminline(A,B) :- a(A), B = #sum { Y: person(A,Y) }.
foo(X) :- X = #sum { F+2,V: suminline(V,F); A: test(A,B) }.
            """,
            [],
            [],
            """#program base.
{ a((1..3)) }.
suminline(A,B) :- a(A); B = #sum { Y: person(A,Y) }.
foo(X) :- X = #sum { (F+2),V: suminline(V,F); A: test(A,B) }.""",
        ),
        (  # replacement in same aggregate but not used in head
            """
{ a((1..3),2) }.
headinline(A,C) :- a(A,C); B = #sum { Y: person(A,Y) }.
foo(X) :- X = #sum { F,V: headinline(V,F); A: test(A,B) }.
            """,
            [],
            [],
            """#program base.
{ a((1..3),2) }.
headinline(A,C) :- a(A,C); B = #sum { Y: person(A,Y) }.
foo(X) :- X = #sum { F,V: headinline(V,F); A: test(A,B) }.""",
        ),
        (  # replacement in same aggregate but used somewhere else
            """
{ a((1..3),2) }.
suminline(A,B) :- a(A,B); B = #sum { Y: person(A,Y) }.
foo(X) :- X = #sum { F,V: suminline(V,F); A: test(A,B) }.
            """,
            [],
            [],
            """#program base.
{ a((1..3),2) }.
suminline(A,B) :- a(A,B); B = #sum { Y: person(A,Y) }.
foo(X) :- X = #sum { F,V: suminline(V,F); A: test(A,B) }.""",
        ),
        (  # replacement in same aggregate, but agg check
            """
{ a((1..3)) }.
suminline(A,B) :- a(A); B = #sum { Y: person(A,Y) } > 13.
foo(X) :- X = #sum { F,V: suminline(V,F); A: test(A,B) }.
            """,
            [],
            [],
            """#program base.
{ a((1..3)) }.
suminline(A,B) :- a(A); B = #sum { Y: person(A,Y) } > 13.
foo(X) :- X = #sum { F,V: suminline(V,F); A: test(A,B) }.""",
        ),
        (  # replacement in same aggregate but negation
            """
{ a((1..3)) }.
suminline(A,B) :- a(A); B = #sum { Y: person(A,Y) }.
foo(X) :- X = #sum { F,V: not suminline(V,F); A: test(A,B) }.
            """,
            [],
            [],
            """#program base.
{ a((1..3)) }.
suminline(A,B) :- a(A); B = #sum { Y: person(A,Y) }.
foo(X) :- X = #sum { F,V: not suminline(V,F); A: test(A,B) }.""",
        ),
        (  # replacement in same aggregate
            """
{ a((1..3)) }.
suminline(A,B) :- a(A); B = #sum { Y: person(A,Y) }.
foo(X) :- X = #sum { F,V: suminline(V,F); A: test(A,B) }.
            """,
            [],
            [],
            """#program base.
{ a((1..3)) }.
foo(X) :- X = #sum { A: test(A,B); Y,V: a(V), person(V,Y) }.""",
        ),
        (  # replacement in same aggregate, several conditions
            """
{ a((1..3)) }.
suminline(A,B) :- a(A), B = #sum { Y,p: person(A,Y); Y,h: human(A,Y) }.
foo(X) :- X = #sum { F,V: suminline(V,F); A: test(A,B) }.
            """,
            [],
            [],
            """#program base.
{ a((1..3)) }.
foo(X) :- X = #sum { A: test(A,B); Y,p,V: a(V), person(V,Y); Y0,h,V: a(V), human(V,Y0) }.""",
        ),
        (  # replacement in same aggregate but negation
            """
{ a((1..3)) }.
neginline(A,B) :- a(A); B = #min {Y : person(A,Y)}.
foo(X) :- X = #min {F,V: not neginline(V,F); A: test(A,B)}.
            """,
            [],
            [],
            """#program base.
{ a((1..3)) }.
neginline(A,B) :- a(A); B = #min { Y: person(A,Y) }.
foo(X) :- X = #min { F,V: not neginline(V,F); A: test(A,B) }.""",
        ),
        (  # replacement in same aggregate
            """
{ a((1..3)) }.
inline(A,B) :- a(A); B = #min {Y : person(A,Y)}.
foo(X) :- X = #min {F,V: inline(V,F); A: test(A,B)}.
            """,
            [],
            [],
            """#program base.
{ a((1..3)) }.
foo(X) :- X = #min { A: test(A,B); Y,V: a(V), person(V,Y) }.""",
        ),
        (  # not replacement in same different
            """
{ a((1..3)) }.
inline(A,B) :- a(A), B = #min {Y : person(A,Y)}.
foo(X) :- X = #sum {F,V: inline(V,F); A: test(A,B)}.
            """,
            [],
            [],
            """#program base.
{ a((1..3)) }.
inline(A,B) :- a(A); B = #min { Y: person(A,Y) }.
foo(X) :- X = #sum { F,V: inline(V,F); A: test(A,B) }.""",
        ),
        (  # replacement in similar aggregate
            """
{ a((1..3)) }.
inline(A,B) :- a(A), B = #count {A,Y : person(A,Y)}.
foo(X) :- X = #sum {F,V: inline(V,F); A: test(A,B)}, bar.
            """,
            [],
            [],
            """#program base.
{ a((1..3)) }.
foo(X) :- X = #sum { A: test(A,B); 1,V,Y,V: a(V), person(V,Y) }; bar.""",
        ),
        (  # replacement in minimize and but no complete inlining because of ambigious set
            """
{ a((1..3)) }.
mininline(A,B) :- a(A), B = #sum {Y : person(A,Y)}.
#minimize {F,V : mininline(V,F); A,B : test(A,B)}.
            """,
            [],
            [],
            """#program base.
{ a((1..3)) }.
:~ a(V); F = #sum { Y: person(V,Y) }. [F@0,V]
:~ test(A,B). [A@0,B]""",
        ),
        (  # replacement in minimize
            """
{ a((1..3)) }.
inline(A,B) :- a(A), B = #sum {Y : person(A,Y)}.
#minimize {F,V: inline(V,F), bar; A,B,test: test(A,B)}.
            """,
            [],
            [],
            """#program base.
{ a((1..3)) }.
:~ a(V); bar; person(V,Y). [Y@0,V,unique,unique]
:~ test(A,B). [A@0,B,test]""",
        ),
        (  # replacement in minimize but input
            """
{ a((1..3)) }.
inline(A,B) :- a(A), B = #sum {Y : person(A,Y)}.
#minimize {F,V: inline(V,F), bar; A,B,test: test(A,B)}.
            """,
            [Predicate("inline", 2)],
            [],
            """#program base.
{ a((1..3)) }.
inline(A,B) :- a(A); B = #sum { Y: person(A,Y) }.
:~ inline(V,F); bar. [F@0,V]
:~ test(A,B). [A@0,B,test]""",
        ),
        (  # replacement in minimize with negation
            """
{ a((1..3)) }.
inline(A,B) :- a(A), B = #sum {Y : person(A,Y)}.
#minimize {F,V: not inline(V,F); A,B,test: test(A,B)}.
            """,
            [],
            [],
            """#program base.
{ a((1..3)) }.
inline(A,B) :- a(A); B = #sum { Y: person(A,Y) }.
:~ not inline(V,F). [F@0,V]
:~ test(A,B). [A@0,B,test]""",
        ),
        (  # replacement in minimize with negation
            """
{ a((1..3)) }.
inline(A,A) :- not a(A).
#minimize {F,V: not inline(V,F); A,B,test: test(A,B)}.
            """,
            [],
            [],
            """#program base.
{ a((1..3)) }.
inline(A,A) :- not a(A).
:~ not inline(V,F). [F@0,V]
:~ test(A,B). [A@0,B,test]""",
        ),
        (  # replacement in minimize with negation
            """
{ a((1..3),2) }.
inline(A,B) :- a(A,B).
#minimize {F,V: not inline(V,F); A,B,test: test(A,B)}.
            """,
            [],
            [],
            """#program base.
{ a((1..3),2) }.
inline(A,B) :- a(A,B).
:~ not inline(V,F). [F@0,V]
:~ test(A,B). [A@0,B,test]""",
        ),
        (  # replacement in minimize with negation
            """
{ a((1..3)) }.
inline(A,A) :- a(A).
#minimize {F,V: inline(V,F); A,B,test: test(A,B)}.
            """,
            [],
            [],
            """#program base.
{ a((1..3)) }.
inline(A,A) :- a(A).
:~ inline(V,F). [F@0,V]
:~ test(A,B). [A@0,B,test]""",
        ),
        (  # replacement in minimize with negation
            """
{ a((1..3)) }.
inline(B) :- B = #sum {Y : person(A,Y)}.
#minimize {F: not inline(F); A,B,test: test(A,B)}.
            """,
            [],
            [],
            """#program base.
{ a((1..3)) }.
inline(B) :- B = #sum { Y: person(A,Y) }.
:~ not inline(F). [F@0]
:~ test(A,B). [A@0,B,test]""",
        ),
        (  # replacement in weak constraint
            """
{ a((1..3)) }.
:~ a(A); B = #sum {Y : person(A,Y)}. [B@0,A]
            """,
            [],
            [],
            """#program base.
{ a((1..3)) }.
:~ a(A); person(A,Y). [Y@0,A,unique]""",
        ),
        (  # replacement in weak constraint, but result is not used as weight
            """
{ a((1..3)) }.
:~ a(A); B = #sum {Y : person(A,Y)}. [A@0,A]
            """,
            [],
            [],
            """#program base.
{ a((1..3)) }.
:~ a(A); B = #sum { Y: person(A,Y) }. [A@0,A]""",
        ),
        (  # replacement in weak constraint, more than 1 agg
            """
{ a((1..3)) }.
:~ a(A); B = #sum {Y : person(A,Y)}; C = #sum {Y : person(A,Y)}. [B@0,A]
            """,
            [],
            [],
            """#program base.
{ a((1..3)) }.
:~ a(A); B = #sum { Y: person(A,Y) }; C = #sum { Y: person(A,Y) }. [B@0,A]""",
        ),
        (  # replacement in weak constraint, but agg has other constraints
            """
{ a((1..3)) }.
:~ a(A); B = #sum {Y : person(A,Y)} > 13. [B@0,A]
            """,
            [],
            [],
            """#program base.
{ a((1..3)) }.
:~ a(A); B = #sum { Y: person(A,Y) } > 13. [B@0,A]""",
        ),
        (  # replacement in weak constraint, but agg is of wrong function
            """
{ a((1..3)) }.
:~ a(A); B = #min {Y : person(A,Y)}. [B@0,A]
            """,
            [],
            [],
            """#program base.
{ a((1..3)) }.
:~ a(A); B = #min { Y: person(A,Y) }. [B@0,A]""",
        ),
        (  # replacement in weak constraint, but result is used
            """
{ a((1..3),2) }.
:~ a(A,B); B = #sum {Y : person(A,Y)}. [B@0,A]
            """,
            [],
            [],
            """#program base.
{ a((1..3),2) }.
:~ a(A,B); B = #sum { Y: person(A,Y) }. [B@0,A]""",
        ),
        (  # replacement in weak constraint, count
            """
{ a((1..3)) }.
:~ a(A); B = #count {Y : person(A,Y)}. [B@0,A]
            """,
            [],
            [],
            """#program base.
{ a((1..3)) }.
:~ a(A); person(A,Y). [1@0,Y,A]""",
        ),
        (  # replacement in combined aggregates
            """
{ mirror_visible(X,Y,GX,GY,D) : dom(X,Y,GX,GY,D) }.
{ direct_visible(X,Y,GX,GY,D) : dom(X,Y,GX,GY,D) }.
ghost(X, Y, N) :- count(X, Y, TN),
         N = #count { GX, GY, D : mirror_visible(X, Y, GX, GY, D), ghost(GX, GY) }.
dracula(X, Y, N) :- count(X, Y, TN),
         N = #count { DX, DY, D : direct_visible(X, Y, DX, DY, D), dracula(DX, DY) }.
:- count(X, Y, N), ghost(X, Y, G), dracula(X, Y, D), zombie(X, Y, Z), N != G + D + Z.
            """,
            [],
            [],
            """#program base.
{ mirror_visible(X,Y,GX,GY,D): dom(X,Y,GX,GY,D) }.
{ direct_visible(X,Y,GX,GY,D): dom(X,Y,GX,GY,D) }.
#false :- count(X,Y,N); count(X,Y,TN);\
 G = #sum+ { 1,GX,GY,D0: mirror_visible(X,Y,GX,GY,D0), ghost(GX,GY) };\
 count(X,Y,TN0);\
 D = #sum+ { 1,DX,DY,D1: direct_visible(X,Y,DX,DY,D1), dracula(DX,DY) };\
 zombie(X,Y,Z); N != ((G+D)+Z).""",
        ),
        (  # not replacement in if not in agg
            """
{ mirror_visible(X,Y,GX,GY,D) : dom(X,Y,GX,GY,D) }.
ghost(X, Y, N) :- count(X, Y, TN),
         N = #count { GX, GY, D : mirror_visible(X, Y, GX, GY, D), ghost(GX, GY) }.
:- count(X, Y, N), ghost(X, Y, G1), dracula(X, Y, D), zombie(X, Y, Z), N != G + D + Z.
            """,
            [],
            [],
            """#program base.
{ mirror_visible(X,Y,GX,GY,D): dom(X,Y,GX,GY,D) }.
ghost(X,Y,N) :- count(X,Y,TN); N = #sum+ { 1,GX,GY,D: mirror_visible(X,Y,GX,GY,D), ghost(GX,GY) }.
#false :- count(X,Y,N); ghost(X,Y,G1); dracula(X,Y,D); zombie(X,Y,Z); N != ((G+D)+Z).""",
        ),
        (  # replacement if directly in agg, lhs
            """
{ mirror_visible(X,Y,GX,GY,D) : dom(X,Y,GX,GY,D) }.
ghost(X, Y, N) :- count(X, Y, TN),
         N = #count { GX, GY, D : mirror_visible(X, Y, GX, GY, D), ghost(GX, GY) }.
:- count(X, Y, N), ghost(X, Y, G1), G1 = #sum { Z,O : zombie(Z,O) }.
            """,
            [],
            [],
            """#program base.
{ mirror_visible(X,Y,GX,GY,D): dom(X,Y,GX,GY,D) }.
#false :- count(X,Y,N);\
 count(X,Y,TN);\
 G1 = #sum+ { 1,GX,GY,D: mirror_visible(X,Y,GX,GY,D), ghost(GX,GY) };\
 G1 = #sum { Z,O: zombie(Z,O) }.""",
        ),
        (  # replacement if directly in agg, rhs
            """
{ mirror_visible(X,Y,GX,GY,D) : dom(X,Y,GX,GY,D) }.
ghost(X, Y, N) :- count(X, Y, TN),
         N = #count { GX, GY, D : mirror_visible(X, Y, GX, GY, D), ghost(GX, GY) }.
:- count(X, Y, N), ghost(X, Y, G1), 0 = #sum { Z,O : zombie(Z,O) } = G1.
            """,
            [],
            [],
            """#program base.
{ mirror_visible(X,Y,GX,GY,D): dom(X,Y,GX,GY,D) }.
#false :- count(X,Y,N);\
 count(X,Y,TN);\
 G1 = #sum+ { 1,GX,GY,D: mirror_visible(X,Y,GX,GY,D), ghost(GX,GY) };\
 0 = #sum { Z,O: zombie(Z,O) } = G1.""",
        ),
        (  # replacement if directly in agg, negation
            """
{ mirror_visible(X,Y,GX,GY,D) : dom(X,Y,GX,GY,D) }.
ghost(N) :- N = #count { GX, GY, D : mirror_visible(X, Y, GX, GY, D), ghost(GX, GY) }.
:- count(X, Y, N), not ghost(G1), 0 = #sum { Z,O : zombie(Z,O) } = G1.
            """,
            [],
            [],
            """#program base.
{ mirror_visible(X,Y,GX,GY,D): dom(X,Y,GX,GY,D) }.
#false :- count(X,Y,N);\
 not G1 = #sum+ { 1,GX,GY,D: mirror_visible(X0,Y0,GX,GY,D), ghost(GX,GY) };\
 0 = #sum { Z,O: zombie(Z,O) } = G1.""",
        ),
        (  # anonymous variables
            """
{ hint(X,Y,Z): dom(X,Y,Z) }.
bend_sums(X,Y,S) :- S = #sum { N,X',Y': hint_path(X',Y',_,_,N,X,Y) }; hint(X,Y,_).
#false :- bend_sums(_,_,S); S != #sum { X : person(X) }.
            """,
            [],
            [],
            """#program base.
{ hint(X,Y,Z): dom(X,Y,Z) }.
bend_sums(X,Y,S) :- S = #sum { N,X',Y': hint_path(X',Y',_,_,N,X,Y) }; hint(X,Y,_).
#false :- bend_sums(_,_,S); S != #sum { X: person(X) }.""",
        ),
    ),
)
def test_inline_translation(
    lhs: str, input_predicates: list[Predicate], output_predicates: list[Predicate], rhs: str
) -> None:
    """test inlining literals"""
    ast: list[AST] = []
    parse_string(lhs, ast.append)
    ast = preprocess(ast)
    utr = InlineTranslator(ast, input_predicates, output_predicates)
    ast = utr.execute(ast)
    ast = postprocess(ast)
    output = "\n".join(map(str, ast))
    assert output == rhs



"""
A module for all predicate dependencies in the AST
"""
from collections import defaultdict
from functools import cache
from itertools import chain, product
from typing import Iterable, Iterator, Mapping

import networkx as nx
from clingo.ast import (
    AST,
    AggregateFunction,
    ASTType,
    BodyAggregate,
    BodyAggregateElement,
    Comparison,
    ComparisonOperator,
    ConditionalLiteral,
    Function,
    Guard,
    Literal,
    Rule,
    Sign,
    SymbolicAtom,
    Variable,
)

from ngo.utils.ast import (
    LOC,
    SIGNS,
    AnnotatedPredicate,
    Predicate,
    SignSetType,
    body_predicates,
    collect_ast,
    collect_bound_variables,
    headderivable_predicates,
    literal_predicate,
    minimize_predicates,
    transform_ast,
)
from ngo.utils.globals import CHAIN_STR, DOM_STR, MAX_STR, MIN_STR, NEXT_STR, UniqueNames


class RuleDependency:
    """get information about heads and body dependencies"""

    def __init__(self, prg: Iterable[AST]):
        self.head2bodies: dict[Predicate, list[AST]] = defaultdict(list)
        self.head2rules: dict[Predicate, list[AST]] = defaultdict(list)
        self.pred2stm: dict[Predicate, list[AST]] = defaultdict(list)

        for stm in prg:
            if stm.ast_type == ASTType.Rule:
                for head in map(
                    lambda x: x.pred,
                    headderivable_predicates(stm),
                ):
                    self.head2bodies[head].append(stm.body)
                    self.head2rules[head].append(stm)
            for p in chain(body_predicates(stm, SIGNS), minimize_predicates(stm, SIGNS)):
                self.pred2stm[p.pred].append(stm)

    def get_bodies(self, head: Predicate) -> list[AST]:
        """return all bodies of head predicate"""
        return self.head2bodies[head]

    def get_rules_that_derive(self, head: Predicate) -> list[AST]:
        """return all rules that could derive head predicate"""
        return self.head2rules[head]

    def get_headderivable_predicates(self) -> list[Predicate]:
        """return all predicates that can possible be derived in a head"""
        return list(self.head2bodies.keys())

    def get_statements_that_use(self, pred: Predicate) -> list[AST]:
        """return all statements that use predicate in body or condition"""
        return self.pred2stm[pred]


# TODO: refactor all graphs
def _create_graph_from_prg(prg: Iterable[AST], signs: SignSetType) -> nx.DiGraph:
    """create a dependency graph from all body predicates (wrt signs) to
    all derivable head predicates
    """
    graph = nx.DiGraph()
    for stm in chain.from_iterable([x.unpool(condition=True) for x in prg]):
        if stm.ast_type == ASTType.Rule:
            heads = headderivable_predicates(stm)
            bodies = body_predicates(stm, signs)
            graph.add_edges_from(
                product(
                    map(lambda signedpred: signedpred.pred, bodies),
                    map(lambda signedpred: signedpred.pred, heads),
                )
            )
    return graph


class DomainPredicates:
    """
    Computes which predicates have static domain and which predicates
    can be used to represent an approximation of the domain.
    Also computes domain predicates, min/max elements and chains.
    """

    def __init__(self, unique_names: UniqueNames, prg: Iterable[AST]):
        self.unique_names = unique_names
        self._not_static: set[Predicate] = set()  # set of predicates that is not already a domain predicate

        prg: list[AST] = list(prg)  # type: ignore
        self.domains: dict[Predicate, Predicate] = {}  # key = ("p",3) -> ("dom",3)
        self.domain_rules: dict[Predicate, list[tuple[AST, list[AST]]]] = defaultdict(list)  # atom -> [conditions, ...]
        self._too_complex: set[
            Predicate
        ] = set()  # set of predicates that is too complex to provide a domain computation
        self.created_domain: set[Predicate] = set()  # set of predicates where I have already created the domain
        self.__compute_nonstatic_predicates(prg)
        self.__compute_domains(prg)

    def __compute_nonstatic_predicates(self, prg: Iterable[AST]) -> None:
        """compate _not_static for all predicates that aren't static"""
        for stm in chain.from_iterable([x.unpool(condition=True) for x in prg]):
            if stm.ast_type == ASTType.Rule:
                ### remove head choice predicates
                head = stm.head
                if head.ast_type in (ASTType.Disjunction, ASTType.Aggregate):
                    for cond in head.elements:
                        assert cond.ast_type == ASTType.ConditionalLiteral
                        lit = list(literal_predicate(cond.literal, SIGNS))[0]
                        self._not_static.add(lit.pred)
                elif head.ast_type == ASTType.HeadAggregate:
                    for elem in filter(lambda x: x.ast_type == ASTType.HeadAggregateElement, head.elements):
                        cond = elem.condition
                        assert cond.ast_type == ASTType.ConditionalLiteral
                        for lit in literal_predicate(cond.literal, SIGNS):
                            self._not_static.add(lit.pred)

        graph = _create_graph_from_prg(prg, SIGNS)
        cycle_free_pdg = graph.copy()
        ### remove predicates in cycles
        for scc in filter(lambda x: len(x) > 1, nx.strongly_connected_components(graph)):
            self._not_static.update(scc)
            cycle_free_pdg.remove_nodes_from(scc)
            self._too_complex.update(scc)
        for scc in nx.selfloop_edges(graph):
            self._not_static.add(scc[0])
            cycle_free_pdg.remove_nodes_from([scc[0]])
            self._too_complex.update([scc[0]])

        ### remove predicates derived by using not_static predicates
        for node in nx.topological_sort(cycle_free_pdg):  # type: ignore
            if any(map(lambda pre: pre in self._not_static, graph.predecessors(node))):
                self._not_static.add(node)

    def is_static(self, pred: Predicate) -> bool:
        """
        returns true if predicate can be computed statically
        """
        return pred not in self._not_static

    def has_domain(self, pred: Predicate) -> bool:
        """
        returns true if a domain of pred has been computed
        """
        return self.is_static(pred) or pred in self.domains

    def domain_predicate(self, pred: Predicate) -> Predicate:
        """
        returns domain predicate of pred
        """
        assert self.has_domain(pred)
        if self.is_static(pred):
            return pred
        return self.domains[pred]

    # important that this is called only once per input.
    # TODO: breaks in multithreading
    @cache  # pylint: disable=method-cache-max-size-none
    def _predicate(self, name: str, arity: int) -> Predicate:
        return self.unique_names.new_predicate(name, arity)

    def min_anon_predicate(self, anon_pred: AnnotatedPredicate, position: int) -> Predicate:
        """
        returns min_domain predicate of annotated pred
        """
        return self._predicate(
            f"{MIN_STR}{'_'.join(str(pos) for pos in anon_pred.annotated_positions)}_{position}"
            + self.domain_predicate(anon_pred.pred).name,
            anon_pred.pred.arity - len(anon_pred.annotated_positions) + 1,
        )

    def max_anon_predicate(self, anon_pred: AnnotatedPredicate, position: int) -> Predicate:
        """
        returns max_domain predicate of annotated pred
        """
        return self._predicate(
            f"{MAX_STR}{'_'.join(str(pos) for pos in anon_pred.annotated_positions)}_{position}"
            + self.domain_predicate(anon_pred.pred).name,
            anon_pred.pred.arity - len(anon_pred.annotated_positions) + 1,
        )

    def next_anon_predicate(self, anon_pred: AnnotatedPredicate, position: int) -> Predicate:
        """
        returns next_domain predicate of annotated pred
        """
        return self._predicate(
            f"{NEXT_STR}{'_'.join(str(pos) for pos in anon_pred.annotated_positions)}_{position}"
            + self.domain_predicate(anon_pred.pred).name,
            anon_pred.pred.arity - len(anon_pred.annotated_positions) + 2,
        )

    def dom_named_predicate(self, name: str, arity: int) -> Predicate:
        """
        returns domain named predicate of name/arity
        """
        return self._predicate(DOM_STR + name, arity)

    def chain_pred(self, anon_pred: AnnotatedPredicate, position: int, maximum: bool) -> Predicate:
        """
        Given an annotated predicate and a position, return the chain predicate
        """
        return self._predicate(
            f"{CHAIN_STR}_{'_'.join(str(pos) for pos in anon_pred.annotated_positions)}_{position}"
            + (MAX_STR if maximum else MIN_STR)
            + self.domain_predicate(anon_pred.pred).name,
            anon_pred.pred.arity - len(anon_pred.annotated_positions) + 1,
        )

    def __create_domain_for_condition(self, node: AST) -> Iterator[AST]:
        """yield all domain rules for all symbolic atoms in node"""
        for symbol in collect_ast(node, "SymbolicAtom"):
            symbol = symbol.symbol
            dom_pred = Predicate(symbol.name, len(symbol.arguments))
            if symbol.ast_type == ASTType.Function:
                orig_pred = [key for key, value in self.domains.items() if value == dom_pred]
                if orig_pred:
                    yield from self.create_domain(orig_pred[0])

    def create_domain(self, pred: Predicate) -> Iterator[AST]:
        """
        given a predicate, yield a list of ast
        that represent rules to create self.domain(pred) in the logic program
        Only yields these rules once, ignoring every subsequent call for the same predicate:
        """
        if pred in self.created_domain:
            return
        self.created_domain.add(pred)
        if not self.has_domain(pred):
            raise RuntimeError(f"Can not create domain for {pred}.")
        if self.is_static(pred):
            return
        if pred not in self.domain_rules:
            raise RuntimeError(f"Can not create domain for {pred}.")  # nocoverage

        for head, condition in self.domain_rules[pred]:
            newatom = SymbolicAtom(
                Function(
                    LOC,
                    self.domain_predicate(pred).name,
                    head.symbol.arguments,
                    False,
                )
            )
            for cond in condition:
                yield from self.__create_domain_for_condition(cond)
            yield Rule(LOC, Literal(LOC, Sign.NoSign, newatom), condition)

    @staticmethod
    def _create_projected_lit(pred: Predicate, variables: Mapping[int, AST], sign: Sign = Sign.NoSign) -> AST:
        """
        Given a predicate pred, create a literal with only projected variables at certain positions.
        Given p/3, {1 : Variable(LOC, "X"), 2: Variable(LOC, "Y")} create Literal p(_,X,Y)
        """
        vars_ = [variables[i] if i in variables else Variable(LOC, "_") for i in range(0, pred.arity)]
        return Literal(
            LOC,
            sign,
            SymbolicAtom(
                Function(
                    LOC,
                    pred.name,
                    vars_,
                    False,
                )
            ),
        )

    @staticmethod
    def _var_map(vars_: list[AST]) -> dict[int, AST]:
        return dict(enumerate(vars_))

    def create_chain_pred_for_annotated_pred(
        self, anon_pred: AnnotatedPredicate, position: int, maximum: bool
    ) -> Iterator[AST]:
        """creating chaining rules from a simple predicate where position points to the arity with the number"""
        pred = anon_pred.pred
        assert position in anon_pred.annotated_positions
        var_p: AST = Variable(LOC, "P")
        var_n: AST = Variable(LOC, "N")
        var_global_map: dict[int, AST] = {
            i: Variable(LOC, f"G{i}") for i in range(0, pred.arity) if i not in anon_pred.annotated_positions
        }
        var_global_flat = [
            Variable(LOC, f"G{i}") for i in range(0, pred.arity) if i not in anon_pred.annotated_positions
        ]
        next_pred = self.next_anon_predicate(anon_pred, position)

        # chain(G1,G2,P) :- orig(G1,G2,_,_,P).
        chain_pred = self.chain_pred(anon_pred, position, maximum)
        yield Rule(
            LOC,
            self._create_projected_lit(chain_pred, self._var_map(var_global_flat + [var_p])),
            [self._create_projected_lit(pred, var_global_map | {position: var_p})],
        )

        prev_agg = var_n
        next_agg = var_p
        if maximum:
            prev_agg = var_p
            next_agg = var_n

        # chain(G1,G2,P) :- chain(G1,G2,L3,L4), __next_dom(G1,G2,P,N).
        yield Rule(
            LOC,
            self._create_projected_lit(chain_pred, self._var_map(var_global_flat + [prev_agg])),
            [
                self._create_projected_lit(chain_pred, self._var_map(var_global_flat + [next_agg])),
                self._create_projected_lit(next_pred, self._var_map(var_global_flat + [var_p, var_n])),
            ],
        )

    def create_next_pred_for_annotated_pred(self, anon_pred: AnnotatedPredicate, position: int) -> Iterator[AST]:
        """
        given a predicate, yield a list of ast
        that represent rules to create a next predicate for the domain of the annotated predicate in the logic program
        includes min/max predicates
        The next predicate is created for the 'position's variable in the predicate, starting with 0
        The annotated variables are assumed to be unique to the non annotated variables
        The position is assumed to be annotated
        """
        # I have global vars, that need to stay the same for the domain next
        #     G1, G2
        # next(G1,G2, 1, 3)
        # next(G1,G2, 3, 10)
        # next(G1,G2, 10, 11)

        # and I also need local Variables that are unique to the single values
        # but not in grounding, so local (annotated) positions may not appear in the next predicates

        # So try to figure out rules to create them
        #     V is the value I try to create the chaining for
        # dom(G1, G2, L1, ..., Ln), 1 < position < n, Lpos=V

        # The global vars are compressed in the next predicates and local variables are simply ommited
        # __min_dom(G1,G2,X) :- X = #min { V: dom(G1,G2,_,_,V) }, dom(G1,G2,_,_,_).
        # __next_dom(G1,G2,P,N) :- __min_dom(G1,G2,P); dom(G1,G2,_,_,P); dom(G1,G2,_,_,N); N > P;
        #                                      not dom(G1,G2,_,_,B): dom(G1,G2,__,B), P < B < N.
        # __next_dom(G1,G2,P,N) :- __next_dom_(G1,G2,_,_,_,_,P); dom(G1,G2,_,_,N); N > P;
        #                                      not dom(G1,G2,_,_,B): dom(G1,G2,__,B), P < B < N.
        # chain(G1,G2,N) :- orig(G1,G2,L1,L2,N).
        # chain(G1,G2,P) :- chain(G1,G2,N), __next_dom(G1,G2,P,N).

        pred = anon_pred.pred

        # TODO: maybe I can combine this function with the old simple function now that it does not use locals anymore ?
        if not self.has_domain(pred):
            raise RuntimeError(f"Can not create order encoding for {pred}. Unable to create domain.")
        if position >= pred.arity:
            raise RuntimeError(
                f"Can not create order encoding for position {position} for {pred}."
                " Position exceeds arity, starting with 0."
            )

        min_pred = self.min_anon_predicate(anon_pred, position)
        max_pred = self.max_anon_predicate(anon_pred, position)
        next_pred = self.next_anon_predicate(anon_pred, position)
        dom_pred = self.domain_predicate(pred)

        var_x: AST = Variable(LOC, "X")
        var_l: AST = Variable(LOC, "L")
        var_p: AST = Variable(LOC, "P")
        var_n: AST = Variable(LOC, "N")
        var_b: AST = Variable(LOC, "B")

        var_global_flat = [
            Variable(LOC, f"G{i}") for i in range(0, pred.arity) if i not in anon_pred.annotated_positions
        ]

        var_global_map: dict[int, AST] = {
            i: Variable(LOC, f"G{i}") for i in range(0, pred.arity) if i not in anon_pred.annotated_positions
        }

        min_body: list[AST] = [
            Literal(
                LOC,
                Sign.NoSign,
                BodyAggregate(
                    LOC,
                    Guard(ComparisonOperator.Equal, var_x),
                    AggregateFunction.Min,
                    [
                        BodyAggregateElement(
                            [var_l], [self._create_projected_lit(dom_pred, var_global_map | {position: var_l})]
                        )
                    ],
                    None,
                ),
            ),
            self._create_projected_lit(dom_pred, var_global_map),
        ]
        ### __min_0__dom_c(G1,G2,X) :- X = #min { L: dom(G1,G2,_,_L) }, dom(G1,G2,_._,_).
        yield Rule(
            LOC,
            self._create_projected_lit(min_pred, self._var_map(var_global_flat + [var_x])),
            min_body,
        )

        max_body: list[AST] = [
            Literal(
                LOC,
                Sign.NoSign,
                BodyAggregate(
                    LOC,
                    Guard(ComparisonOperator.Equal, var_x),
                    AggregateFunction.Max,
                    [
                        BodyAggregateElement(
                            [var_l], [self._create_projected_lit(dom_pred, var_global_map | {position: var_l})]
                        )
                    ],
                    None,
                ),
            ),
            self._create_projected_lit(dom_pred, var_global_map),
        ]
        ### __max_0__dom_c(G1,G2,X) :- X = #max { L: dom(G1,G2,_,_L) }, dom(G1,G2,_._,_).
        yield Rule(
            LOC,
            self._create_projected_lit(max_pred, self._var_map(var_global_flat + [var_x])),
            max_body,
        )

        var_all_p: list[AST] = []
        for i in range(anon_pred.pred.arity):
            if i not in anon_pred.annotated_positions:
                var_all_p.append(Variable(LOC, f"G{i}"))
            else:
                if i == position:
                    var_all_p.append(var_p)
                else:
                    var_all_p.append(Variable(LOC, f"L{i}"))

        # __next_dom(G1,G2,P,N) :- __min_dom(G1,G2,P); dom(G1,G2,N); N > P;
        #                                      not dom(G1,G2,_,_,B): dom(G1,G2,__,B), P < B < N.

        yield Rule(
            LOC,
            self._create_projected_lit(next_pred, self._var_map(var_global_flat + [var_p, var_n])),
            [
                self._create_projected_lit(min_pred, self._var_map(var_global_flat + [var_p])),
                self._create_projected_lit(dom_pred, var_global_map | {position: var_n}),
                Literal(
                    LOC,
                    Sign.NoSign,
                    Comparison(var_n, [Guard(ComparisonOperator.GreaterThan, var_p)]),
                ),
                ConditionalLiteral(
                    LOC,
                    self._create_projected_lit(dom_pred, var_global_map | {position: var_b}, Sign.Negation),
                    [
                        self._create_projected_lit(dom_pred, var_global_map | {position: var_b}),
                        Literal(
                            LOC,
                            Sign.NoSign,
                            Comparison(
                                var_p,
                                [
                                    Guard(ComparisonOperator.LessThan, var_b),
                                    Guard(ComparisonOperator.LessThan, var_n),
                                ],
                            ),
                        ),
                    ],
                ),
            ],
        )

        # __next_dom(G1,G2,P,N) :- __next_dom_(G1,G2,_,P); dom(G1,G2,_,_,N); N > P;
        #                                      not dom(G1,G2,_,_,B): dom(G1,G2,__,B), P < B < N.

        yield Rule(
            LOC,
            self._create_projected_lit(next_pred, self._var_map(var_global_flat + [var_p, var_n])),
            [
                self._create_projected_lit(
                    next_pred,
                    self._var_map(var_global_flat + [Variable(LOC, "_"), var_p]),
                ),
                self._create_projected_lit(dom_pred, var_global_map | {position: var_n}),
                Literal(
                    LOC,
                    Sign.NoSign,
                    Comparison(var_n, [Guard(ComparisonOperator.GreaterThan, var_p)]),
                ),
                ConditionalLiteral(
                    LOC,
                    self._create_projected_lit(dom_pred, var_global_map | {position: var_b}, Sign.Negation),
                    [
                        self._create_projected_lit(dom_pred, var_global_map | {position: var_b}),
                        Literal(
                            LOC,
                            Sign.NoSign,
                            Comparison(
                                var_p,
                                [
                                    Guard(ComparisonOperator.LessThan, var_b),
                                    Guard(ComparisonOperator.LessThan, var_n),
                                ],
                            ),
                        ),
                    ],
                ),
            ],
        )

    def add_domain_rule(self, pred: Predicate, conditions: list[tuple[AST, list[AST]]]) -> None:
        """forward to add_domain_rules and make pred non static"""
        domain_rules: dict[Predicate, list[tuple[AST, list[AST]]]] = {}
        domain_rules[pred] = conditions
        self._not_static.add(pred)
        self.add_domain_rules(domain_rules)

    def add_domain_rules(self, domain_rules: dict[Predicate, list[tuple[AST, list[AST]]]]) -> None:
        """
        add atom <- body[0] or body[1] ... body[n] to self.domain_rules
        if it passes all checks to compute an actual domain
        inserts elements in the body as their domains
        also mark the head as not static
        """
        # for pred in domain_rules
        # for atom in domain_rules:
        #     if atom.ast_type == ASTType.SymbolicAtom:
        #         self._not_static.update([Predicate(atom.symbol.name, len(atom.symbol.arguments))])
        # TODO: refactor, no intermediate map needed anymore, could work on self.domain_rules
        # refactoring then only needs to take new domain rules into account

        def is_too_complex(cond: AST) -> bool:
            for atom in collect_ast(cond, "SymbolicAtom"):
                name = atom.symbol.name
                arity = len(atom.symbol.arguments)
                if Predicate(name, arity) in self._too_complex:
                    return True
            return False

        # ### filter out conditions that can not be translated to domain conditions
        # ### like a sum calculation
        def is_dynamic_sum(cond: AST) -> bool:
            if cond.ast_type != ASTType.Literal:  # maybe conditional literal
                return False
            cond = cond.atom
            if cond.ast_type in (ASTType.BodyAggregate, ASTType.Aggregate):
                for elem in cond.elements:
                    for atom in collect_ast(elem, "SymbolicAtom"):
                        name = atom.symbol.name
                        arity = len(atom.symbol.arguments)
                        if not self.is_static(Predicate(name, arity)):
                            return True
            return False

        def unbounded_head(pair: tuple[AST, list[AST]]) -> bool:
            (head, condition) = pair
            head_variables: set[str] = set(map(lambda x: x.name, collect_ast(head, "Variable")))
            head_variables -= set(map(lambda x: x.name, collect_bound_variables(condition)))
            if len(head_variables) > 0:
                return True
            return False

        def too_complex_rules(pair: tuple[Predicate, list[tuple[AST, list[AST]]]]) -> bool:
            pred, rules = pair
            if self.is_static(pred):
                return True
            for rule in rules:
                if unbounded_head(rule):
                    return True
                (head, condition) = rule
                if is_too_complex(head):
                    return True
                for cond in condition:
                    if is_too_complex(cond) or is_dynamic_sum(cond):
                        return True
            return False

        domain_rules = dict(filter(lambda x: not too_complex_rules(x), domain_rules.items()))

        def have_domain(lit: AST) -> bool:
            for atom in collect_ast(lit, "SymbolicAtom"):
                if atom.symbol.ast_type == ASTType.Function:
                    if not self.has_domain(Predicate(atom.symbol.name, len(atom.symbol.arguments))):
                        return False
            return True

        def replace_domain(atom: AST) -> AST:
            assert atom.ast_type == ASTType.SymbolicAtom
            assert atom.symbol.ast_type == ASTType.Function  # not necessary, but I still have to handle the case, TODO
            name = atom.symbol.name
            arity = len(atom.symbol.arguments)
            assert self.has_domain(Predicate(name, arity))
            return atom.update(symbol=atom.symbol.update(name=self.domain_predicate(Predicate(name, arity)).name))

        while True:
            num_domain_preds = len(self.domain_rules)
            for pred, rules in domain_rules.items():
                if pred in self.domains:
                    continue  # we already computed the domain
                # check if all rules have domain versions
                for rule in rules:
                    head, condition = rule
                    if not all(map(have_domain, condition)):
                        break
                else:
                    # replace all predicates with their respective domain predicates
                    new_rules: list[tuple[AST, list[AST]]] = []
                    for rule in rules:
                        head, condition = rule
                        new_condition: list[AST] = []
                        for cond in condition:
                            new_condition.append(transform_ast(cond, "SymbolicAtom", replace_domain))
                        new_rules.append((head, new_condition))
                    self.domain_rules[pred] = new_rules
                    self.domains[pred] = self.dom_named_predicate(pred.name, pred.arity)
                    self._not_static.add(pred)

            if len(self.domain_rules) == num_domain_preds:
                break

    def __compute_domains(self, prg: Iterable[AST]) -> None:
        """compute self.domain_rules with atom as key and a list of conditions"""
        domain_rules: dict[Predicate, list[tuple[AST, list[AST]]]] = defaultdict(list)

        def atom2pred(atom: AST) -> Predicate:
            assert atom.ast_type == ASTType.SymbolicAtom
            return Predicate(atom.symbol.name, len(atom.symbol.arguments))

        ### collect conditions for the head
        for rule in filter(
            lambda rule: rule.ast_type == ASTType.Rule, chain.from_iterable([x.unpool(condition=True) for x in prg])
        ):
            head = rule.head
            body = rule.body
            if (
                head.ast_type == ASTType.Literal
                and head.sign == Sign.NoSign
                and head.atom.ast_type == ASTType.SymbolicAtom
            ):
                domain_rules[atom2pred(head.atom)].append((head.atom, body))
            elif head.ast_type == ASTType.Disjunction:
                for elem in head.elements:
                    assert elem.ast_type == ASTType.ConditionalLiteral
                    condition = elem.condition
                    if elem.literal.sign == Sign.NoSign and elem.literal.atom.ast_type == ASTType.SymbolicAtom:
                        domain_rules[atom2pred(elem.literal.atom)].append(
                            (elem.literal.atom, list(chain(condition, body)))
                        )
            elif head.ast_type == ASTType.HeadAggregate:
                for elem in filter(
                    lambda elem: elem.condition.literal.sign == Sign.NoSign
                    and elem.condition.literal.atom.ast_type == ASTType.SymbolicAtom,
                    head.elements,
                ):
                    domain_rules[atom2pred(elem.condition.literal.atom)].append(
                        (elem.condition.literal.atom, list(chain(elem.condition.condition, body)))
                    )
            elif head.ast_type == ASTType.Aggregate:
                for elem in filter(
                    lambda elem: elem.literal.sign == Sign.NoSign
                    and elem.literal.atom.ast_type == ASTType.SymbolicAtom,
                    head.elements,
                ):
                    domain_rules[atom2pred(elem.literal.atom)].append(
                        (elem.literal.atom, list(chain(elem.condition, body)))
                    )
        self.add_domain_rules(domain_rules)



""" testing the dependency graph calculations """
import pytest
from clingo.ast import AST, Sign, Transformer, parse_string

from ngo.dependency import DomainPredicates
from ngo.normalize import postprocess, preprocess
from ngo.utils.ast import AnnotatedPredicate, Predicate, SignedPredicate, body_predicates, head_predicates
from ngo.utils.globals import UniqueNames


@pytest.mark.parametrize(
    "rule, result",
    [
        (
            "#false :- 1 <= #sum { 1,a: a; 1,b: b; 1,c: c } <= 2,"
            " X = #sum { 1,e: e; 1,f: f; 1,g: g } 3, X>=2, 5>3, X=Y, 1<=X!=4<5.",
            [
                SignedPredicate(Sign.NoSign, Predicate("a", 0)),
                SignedPredicate(Sign.NoSign, Predicate("b", 0)),
                SignedPredicate(Sign.NoSign, Predicate("c", 0)),
                SignedPredicate(Sign.NoSign, Predicate("e", 0)),
                SignedPredicate(Sign.NoSign, Predicate("f", 0)),
                SignedPredicate(Sign.NoSign, Predicate("g", 0)),
            ],
        ),
        (
            "#false :- 1 { a : e; b : not f; c } 2, d.",
            [
                SignedPredicate(Sign.NoSign, Predicate("a", 0)),
                SignedPredicate(Sign.NoSign, Predicate("b", 0)),
                SignedPredicate(Sign.NoSign, Predicate("c", 0)),
                SignedPredicate(Sign.NoSign, Predicate("d", 0)),
                SignedPredicate(Sign.NoSign, Predicate("e", 0)),
            ],
        ),
    ],
)
def test_positive_body(rule: str, result: list[SignedPredicate]) -> None:
    """test the collection of predicates in positive body context"""

    class RuleVisitor(Transformer):
        """Simple Transformer"""

        def visit_Rule(self, stm: AST) -> AST:  # pylint: disable=invalid-name
            """derived visit method"""
            assert set(body_predicates(stm, {Sign.NoSign})) == set(result)
            return stm

    ruler = RuleVisitor()
    parse_string(rule, ruler)  # type: ignore


@pytest.mark.parametrize(
    "rule, result",
    [
        (
            "a(1,4,f(4)).",
            [SignedPredicate(Sign.NoSign, Predicate("a", 3))],
        ),
        (
            "1 <= #sum { 1,a: a; 1,b: b; 1: c } <= 2.",
            [
                SignedPredicate(Sign.NoSign, Predicate("a", 0)),
                SignedPredicate(Sign.NoSign, Predicate("b", 0)),
                SignedPredicate(Sign.NoSign, Predicate("c", 0)),
            ],
        ),
        (
            "1 { a : e; b : not f; c } 2.",
            [
                SignedPredicate(Sign.NoSign, Predicate("a", 0)),
                SignedPredicate(Sign.NoSign, Predicate("b", 0)),
                SignedPredicate(Sign.NoSign, Predicate("c", 0)),
                SignedPredicate(Sign.NoSign, Predicate("e", 0)),
            ],
        ),
        (
            "a; b; not c.",
            [SignedPredicate(Sign.NoSign, Predicate("a", 0)), SignedPredicate(Sign.NoSign, Predicate("b", 0))],
        ),
        (
            "a : d; b : not e; not c.",
            [
                SignedPredicate(Sign.NoSign, Predicate("a", 0)),
                SignedPredicate(Sign.NoSign, Predicate("b", 0)),
                SignedPredicate(Sign.NoSign, Predicate("d", 0)),
            ],
        ),
    ],
)
def test_positive_head(rule: str, result: list[SignedPredicate]) -> None:
    """test the collection of predicates in positive head context"""

    class RuleVisitor(Transformer):
        """Simple Transformer"""

        def visit_Rule(self, stm: AST) -> AST:  # pylint: disable=invalid-name
            """derived visit method"""
            assert set(head_predicates(stm, {Sign.NoSign})) == set(result)
            return stm

    ruler = RuleVisitor()
    parse_string(rule, ruler)  # type: ignore


@pytest.mark.parametrize(
    "prg, static, notstatic, hasdomain",
    [
        (
            """
            b :- a.
            c :- b.
            d :- c.
            a :- d.
            e :- d.
            """,
            [Predicate("x", 3)],
            [
                Predicate("a", 0),
                Predicate("b", 0),
                Predicate("c", 0),
                Predicate("d", 0),
                Predicate("e", 0),
            ],
            [
                Predicate("x", 3),
            ],
        ),
        (
            """
            b :- #sum{1 : a}.
            c :- b.
            {d} :- c, not d.
            a :- d, not e.
            e :- d.
            f;g :- e.
            e;not g :- f.
            e :- not g.
            x :- not x.
            y.
            z :- y.
            {w}.
            u ; v.
            p(1) :- w.
            q(X) :- not p(X).
            """,
            [
                Predicate("y", 0),
                Predicate("z", 0),
            ],
            [
                Predicate("a", 0),
                Predicate("b", 0),
                Predicate("c", 0),
                Predicate("d", 0),
                Predicate("e", 0),
                Predicate("f", 0),
                Predicate("g", 0),
                Predicate("x", 0),
                Predicate("w", 0),
                Predicate("u", 0),
                Predicate("v", 0),
                Predicate("p", 1),
                Predicate("q", 1),
            ],
            [
                Predicate("y", 0),
                Predicate("z", 0),
                Predicate("w", 0),
                Predicate("u", 0),
                Predicate("v", 0),
                Predicate("p", 1),
            ],
        ),
        (
            """
            a(X) :- b(X,Y), c(Y).
            {d(X)} :- b(X,Y), c(Y).
            """,
            [
                Predicate("a", 1),
                Predicate("b", 2),
                Predicate("c", 1),
            ],
            [Predicate("d", 1)],
            [
                Predicate("a", 1),
                Predicate("b", 2),
                Predicate("c", 1),
                Predicate("d", 1),
            ],
        ),
        (
            """
            a(X,Y) :- a(Y,X).
            a(A,B) :- start(A,B).
            """,
            [],
            [
                Predicate("a", 2),
            ],
            [],
        ),
        (
            """
            1 = #sum {1,a : a(X)} :- b(X).
            """,
            [Predicate("b", 1)],
            [
                Predicate("a", 1),
            ],
            [
                Predicate("a", 1),
            ],
        ),
        (
            """
            {c(X,X) : b(X)}.
            a(X) :- c(X,Y) : b(Y).
            """,
            [Predicate("b", 1)],
            [
                Predicate("c", 2),
                Predicate("a", 1),
            ],
            [Predicate("c", 2)],
        ),
        (
            """
            1 = { path(X1,Y1,A,B): cell(A,B) } :- cell(X1,Y1); not final(X1,Y1).
            """,
            [Predicate("cell", 2)],
            [
                Predicate("path", 4),
            ],
            [Predicate("cell", 2)],
        ),
        (
            """
            a(Z) :- Z = Y, c(Y).
            """,
            [
                Predicate("a", 1),
                Predicate("c", 1),
            ],
            [],
            [
                Predicate("a", 1),
                Predicate("c", 1),
            ],
        ),
    ],
)
def test_domain_predicates(
    prg: str, static: list[Predicate], notstatic: list[Predicate], hasdomain: list[Predicate]
) -> None:
    """test computation of static/non-static predicates and
    non cyclic domain inference"""
    ast: list[AST] = []
    parse_string(prg, ast.append)
    ast = preprocess(ast)
    unique = UniqueNames(ast, [])
    dp = DomainPredicates(unique, ast)
    for pred in static:
        assert dp.is_static(pred)
    for pred in notstatic:
        assert not dp.is_static(pred)
    for pred in hasdomain:
        assert dp.has_domain(pred)


@pytest.mark.parametrize(
    "prg, hasnodomain",
    [
        (
            """
            a(X,Y) :- a(Y,X).
            a(A,B) :- start(A,B).
            """,
            [
                Predicate("a", 2),
            ],
        ),
        (
            """
            a(X,Y) :- a(Y,X).
            b(A,B) :- a(A,B).
            """,
            [
                Predicate("a", 2),
            ],
        ),
        (
            """
            1 = { at((X+A),(Y+B),S): field((X+A),(Y+B)), diff(A,B), not wall(X,Y,(X+A),(Y+B)) } :-
                 step(S); at(X,Y,(S-1)); goal(X1,Y1); not at(X1,Y1,(S-1)).
            at(X,Y,0) :- start(X,Y).
            """,
            [
                Predicate("at", 3),
            ],
        ),
        (
            """
            1 = #sum {1,a : a(X)} :- a(X).
            """,
            [
                Predicate("a", 1),
            ],
        ),
    ],
)
def test_nodomain_predicates(prg: str, hasnodomain: list[Predicate]) -> None:
    """test computation of non cyclic domain inference"""
    ast: list[AST] = []
    parse_string(prg, ast.append)
    ast = preprocess(ast)
    unique = UniqueNames(ast, [])
    dp = DomainPredicates(unique, ast)
    for pred in hasnodomain:
        assert not dp.has_domain(pred)


@pytest.mark.parametrize(
    "prg, predicates, domain_program",
    [
        (
            """
            a(X) :- b(X,Y), c(Y).
            {d(X)} :- b(X,Y), c(Y).
            e(X) :- a(X).
            {f(X)} :- d(X), a(X).
            {g(X)} :- d(X), a(Y), X <= Y.
            h(a(X), X+1, 42) :- g(X), not f(X).
            i(4).
            {i(X)} :- a(X).
            {j(X)} :- a(X).
            {j(Y) : b(X,Y)}.
            {k(Y)} :- Y=X+1, a(X).
            {l(Y)} :- Y=X+1, l(X), Y < 100.
            """,
            [
                Predicate("d", 1),
                Predicate("f", 1),
                Predicate("g", 1),
                Predicate("h", 3),
                Predicate("i", 1),
                Predicate("j", 1),
                Predicate("k", 1),
                Predicate("l", 1),
            ],
            [
                "__dom_d(X) :- b(X,Y); c(Y).",
                "__dom_f(X) :- __dom_d(X); a(X).",
                "__dom_g(X) :- __dom_d(X); a(Y); X <= Y.",
                "__dom_h(a(X),(X+1),42) :- __dom_g(X); not __dom_f(X).",
                "__dom_i(4).",
                "__dom_i(X) :- a(X).",
                "__dom_j(X) :- a(X).",
                "__dom_j(Y) :- b(X,Y).",
                "__dom_k((X+1)) :- a(X).",
            ],
        ),
        (
            """
            {b(Y) : a(Y)}.
            {c(X)} :- X = #sum {1, Y: b(Y)}.
            """,
            [
                Predicate("a", 1),
                Predicate("b", 1),
                Predicate("c", 1),
            ],
            [
                "__dom_b(Y) :- a(Y).",
            ],
        ),
        (
            """
            {l(Y)} :- Y=X+1, l(X), Y < 100.
            a(X) :- l(X).
            """,
            [
                Predicate("l", 1),
                Predicate("a", 1),
                Predicate("c", 1),
            ],
            [],
        ),
        (
            """
            {person(a);
            person(b)}.
            {
            skill(a, ("some",1), 3);
            skill(a, ("thing",1), 5);
            skill(a, ("programming",1..10), 10);
            skill(a, ("knitting",1..10), 100);
            skill(b, t("cooking",1..10), 10);
            skill(b, t("knitting",1..10), 1)}.
            max(P, X) :- X = #max {V, ID : skill(P, ID, V)}, person(P).
            """,
            [
                Predicate("max", 2),
            ],
            [],
        ),
        (
            """
            {person(a);
            person(b)}.
            max(P, X) :- X = #max {V, ID : skull(P, ID, V)}, person(P).
            """,
            [
                Predicate("max", 2),
            ],
            [
                "__dom_person(a).",
                "__dom_person(b).",
                "__dom_max(P,X) :- X = #max { V,ID: skull(P,ID,V) }; __dom_person(P).",
            ],
        ),
        (
            """
            {person(a);
            person(b)}.
            a(X) :- b(X).
            b(X) :- a(X).
            {c(X) : b(X) } :- b(X).
            """,
            [
                Predicate("person", 1),
                Predicate("a", 1),
                Predicate("b", 1),
                Predicate("c", 1),
            ],
            [
                "__dom_person(a).",
                "__dom_person(b).",
            ],
        ),
        (
            """
            a(X) | b(X) :- c(X).
            """,
            [
                Predicate("a", 1),
                Predicate("b", 1),
                Predicate("c", 1),
            ],
            [
                "__dom_a(X) :- c(X).",
                "__dom_b(X) :- c(X).",
            ],
        ),
        (
            """
            {c(X,X) : b(X)}.
            a(X) :- c(X,Y) : b(Y).
            """,
            [
                Predicate("a", 1),
                Predicate("b", 1),
                Predicate("c", 2),
            ],
            [
                "__dom_c(X,X) :- b(X).",
            ],
        ),
        (
            """
            {b(X) : a(X)}.
            c(X) :- X = #max{X : b(X)}.
            d(X) :- c(X).
            """,
            [
                Predicate("a", 1),
                Predicate("b", 1),
                Predicate("c", 1),
                Predicate("d", 1),
            ],
            [
                "__dom_b(X) :- a(X).",
            ],
        ),
        (
            """
            1 = { at((X+A),(Y+B),S): field((X+A),(Y+B)), diff(A,B), not wall(X,Y,(X+A),(Y+B)) } :-
              step(S); at(X,Y,(S-1)); goal(X1,Y1); not at(X1,Y1,(S-1)).
            end(S) :- goal(X,Y); at(X,Y,S).
            """,
            [
                Predicate("end", 1),
            ],
            [],
        ),
        (
            """
            {a(Z)} :- Y = Z, c(Y).
            """,
            [
                Predicate("a", 1),
                Predicate("c", 1),
            ],
            [
                "__dom_a(Z) :- c(Z).",
            ],
        ),
        (
            """
            {c(X,X) : b(X)}.
            """,
            [
                Predicate("b", 1),
                Predicate("c", 2),
            ],
            [
                "__dom_c(X,X) :- b(X).",
            ],
        ),
        (
            """
            {c(1..2, 3..4)}.
            a(X) :- d(X), c(X,Y) : b(Y).
            """,
            [
                Predicate("a", 1),
                Predicate("c", 2),
            ],
            ["__dom_c((1..2),(3..4)).", "__dom_a(X) :- d(X); __dom_c(X,Y): b(Y)."],
        ),
        (
            """
            {c(1..2, 3..4)}.
            {b((1..4))}.
            a(X) :- d(X); c(X,Y) : b(Y).
            """,
            [
                Predicate("a", 1),
                Predicate("b", 1),
                Predicate("c", 2),
            ],
            ["__dom_c((1..2),(3..4)).", "__dom_b((1..4)).", "__dom_a(X) :- d(X); __dom_c(X,Y): __dom_b(Y)."],
        ),
        (
            """
size(S): S = (N*N) :- subgrid_size(N).
            """,
            [
                Predicate("size", 1),
            ],
            ["__dom_size((N*N)) :- subgrid_size(N)."],
        ),
    ],
)
def test_domain_predicates_condition(prg: str, predicates: list[Predicate], domain_program: list[str]) -> None:
    """test domain computation on whole programs"""
    ast: list[AST] = []
    parse_string(prg, ast.append)
    unique = UniqueNames(ast, [])
    ast = preprocess(ast)
    dp = DomainPredicates(unique, ast)
    strlist: list[str] = []
    for pred in predicates:
        if dp.has_domain(pred):
            strlist.extend(map(str, postprocess(dp.create_domain(pred))))
    assert sorted(strlist) == sorted(domain_program)


def test_domain_predicates_exceptions() -> None:
    # pragma: no cover
    """test domain computation exceptions"""
    ast: list[AST] = []
    parse_string("a(X) :- b(X). b(X) :- a(X).", ast.append)
    ast = preprocess(ast)
    with pytest.raises(Exception):
        unique = UniqueNames(ast, [])
        dp = DomainPredicates(unique, ast)
        list(dp.create_domain(Predicate("a", 1)))

    with pytest.raises(Exception):
        unique = UniqueNames(ast, [])
        dp = DomainPredicates(unique, ast)
        list(dp.create_next_pred_for_annotated_pred(AnnotatedPredicate(Predicate("a", 1), (0,)), 0))

    with pytest.raises(Exception):
        unique = UniqueNames(ast, [])
        dp = DomainPredicates(unique, ast)
        list(dp.create_next_pred_for_annotated_pred(AnnotatedPredicate(Predicate("c", 1), (0,)), 1))



"""
The command line parser for the project.
"""

import sys
from argparse import Action, ArgumentParser, ArgumentTypeError, Namespace
from textwrap import dedent
from typing import Any, Optional

from ngo.utils.ast import Predicate

__all__ = ["get_parser"]

if sys.version_info[1] < 8:
    import importlib_metadata as metadata  # nocoverage
else:
    from importlib import metadata  # nocoverage

VERSION = metadata.version("ngo")


ALL_OPTIONS = [
    "minmax_chains",
    "symmetry",
    "duplication",
    "cleanup",
    "unused",
    "sum_chains",
    "math",
    "inline",
    "projection",
]
DEFAULT_OPTIONS = ["minmax_chains", "symmetry", "cleanup", "unused", "sum_chains", "math", "inline", "projection"]

LEVELS = ["ERROR", "WARNING", "INFO", "DEBUG"]


class PredicateList(Action):
    """
    parse a list of input predicates or call auto detection
    """

    def __call__(
        self, parser: ArgumentParser, namespace: Namespace, values: Any, option_string: Optional[str] = None
    ) -> None:
        if values == "auto":
            setattr(namespace, self.dest, values)
        elif values is None or values == "":
            setattr(namespace, self.dest, [])
        else:
            pred_list: list[Predicate] = []
            predicate_strings = values.split(",")
            for predicate_string in predicate_strings:
                sl = predicate_string.split("/")
                if len(sl) != 2:
                    raise ArgumentTypeError(
                        f"Predicates have to be of the form name/arity, but got {predicate_string}."
                    )
                name = sl[0].strip(" ")
                try:
                    arity = int(sl[1])
                except ValueError as esc:
                    raise ArgumentTypeError(f"Arity of predicate {name} should be integer, but got {sl[1]}.") from esc
                pred_list.append(Predicate(name, arity))
            setattr(namespace, self.dest, pred_list)


class VerifyEnable(Action):
    """
    verify and extend enable option set
    """

    def __call__(
        self, parser: ArgumentParser, namespace: Namespace, values: Any, option_string: Optional[str] = None
    ) -> None:
        if len(values) > 1 and "none" in values:
            raise ArgumentTypeError("'none' may not be combined with other options.")
        if "all" in values:
            values = ALL_OPTIONS

        if "default" in values:
            while "default" in values:
                values.remove("default")
            values = sorted(values + DEFAULT_OPTIONS)

        setattr(namespace, self.dest, values)


def get_parser() -> ArgumentParser:
    """
    Return the parser for command line options.
    """
    parser = ArgumentParser(
        prog="ngo",
        description=dedent(
            """\
            ngo
            non ground optimizer
            """
        ),
    )

    parser.add_argument(
        "--log",
        default="info",
        choices=LEVELS,
        type=str.upper,
        metavar=f"{{{','.join(LEVELS)}}}",
        help="set log level [%(default)s]",
    )

    parser.add_argument(
        "--enable",
        action=VerifyEnable,
        type=str.lower,
        nargs="+",
        choices=["all", "none", "default"] + ALL_OPTIONS,
        default=DEFAULT_OPTIONS,
        help=f"enables a set of traits, or default={' '.join(DEFAULT_OPTIONS)}",
    )
    parser.add_argument(
        "--input-predicates",
        action=PredicateList,
        nargs="?",
        type=str,
        default="auto",
        help="enter all input predicates in the form 'name/arity', like 'edge/2' as a comma seperated list.",
    )
    parser.add_argument(
        "--output-predicates",
        action=PredicateList,
        nargs="?",
        type=str,
        default="auto",
        help="enter all output predicates in the form 'name/arity', like 'edge/2' as a comma seperated list.",
    )

    parser.add_argument("--version", "-v", action="version", version=f"%(prog)s {VERSION}")
    return parser



""" test removal of superseeded literals """
import pytest
from clingo.ast import AST, parse_string

from ngo.api import optimize
from ngo.normalize import preprocess
from ngo.utils.globals import auto_detect_input


@pytest.mark.parametrize(
    "lhs, rhs",
    (
        (
            """
dur(J,M,D) :- duration(J,M,D).
job(J) :- dur(J,_,_).
machine(M) :- dur(_,M,_).
jobs(X) :- X = { job(J) }.
machines(M) :- machine(M); not machine((M+1)).
time(T) :- T = (1..X); X = #sum { D,J,M: dur(J,M,D)}.
{ slot(J,M,T): duration(J,M,_) } :- machine(M); time(T).
:- slot(J,M,T), not slot(J,M,T-1), duration(J,M,D),     time(T+D), X=T..T+D-1, not slot(J,M,X).
:- slot(J1,M,T); slot(J2,M,T); J1 != J2; machine(M).
:- duration(J,M,T), not slot(J,M,_).
:- sequence(J,M,S), sequence(J,MM,(S-1)), slot(J,M,T), slot(J,MM,TT), TT >= T.
all_finish(Max) :- Max = #max {T : slot(J,M,T) }.
#minimize {T : all_finish(T)}.
""",
            """#program base.
machine(M) :- duration(_,M,_).
time(T) :- T = (1..X); X = #sum { D,J,M: duration(J,M,D) }.
{ slot(J,M,T): duration(J,M,_) } :- machine(M); time(T).
#false :- slot(J,M,T); not slot(J,M,(T-1)); duration(J,M,D); time((T+D)); X = (T..((T+D)-1)); not slot(J,M,X).
__aux_1(__AUX_1) :- duration(_,__AUX_1,_); machine(__AUX_1).
__dom_slot(M,T) :- time(T); __aux_1(M).
#false :- __dom_slot(M,T); 2 <= #count { J1: slot(J1,M,T) }.
#false :- duration(J,M,_); not slot(J,M,_).
#false :- sequence(J,M,S); sequence(J,MM,(S-1)); slot(J,M,T); slot(J,MM,TT); TT >= T.
__dom_slot1(T) :- time(T); __aux_1(_).
__min_0_0__dom___max_0_13(X) :- X = #min { L: __dom_slot1(L) }; __dom_slot1(_).
__next_0_0__dom___max_0_13(P,N) :- __min_0_0__dom___max_0_13(P); __dom_slot1(N); N > P;\
 not __dom_slot1(B): __dom_slot1(B), P < B < N.
__next_0_0__dom___max_0_13(P,N) :- __next_0_0__dom___max_0_13(_,P); __dom_slot1(N); N > P;\
 not __dom_slot1(B): __dom_slot1(B), P < B < N.
__chain_0_0__max___dom___max_0_13(T) :- slot(_,_,T).
__aux_2(__AUX_0,__AUX_1) :- __chain_0_0__max___dom___max_0_13(__AUX_0); __next_0_0__dom___max_0_13(__AUX_1,__AUX_0).
__chain_0_0__max___dom___max_0_13(__PREV) :- __aux_2(_,__PREV).
:~ __aux_2(__NEXT,__PREV). [(__NEXT-__PREV)@0,__chain_0_0__max___dom___max_0_13(__PREV,__NEXT)]
:~ __chain_0_0__max___dom___max_0_13(__NEXT);\
 __min_0_0__dom___max_0_13(__NEXT). [__NEXT@0,__chain_0_0__max___dom___max_0_13(#sup,__NEXT)]""",
        ),
        (
            """
{max(P, X)} :- X = #max {V, ID : P=42, skill(P, ID, V); 23 : #true}, person(P), random(Y).
            """,  # currently not handled but in future, see #9
            """#program base.
{ max(P,X) } :- X = #max { V,ID: P = 42, skill(P,ID,V); 23 }; person(P); random(_).""",
        ),
        (
            """
subgrid_size(3).
size(S): S = (N*N) :- subgrid_size(N).
sudoku(X,Y,V) :- initial(X,Y,V).
1 = { sudoku(X,Y,V): V = (1..N) } :- size(N); X = (1..N); Y = (1..N).
#false :- sudoku(X1,Y1,V1); sudoku(X2,Y2,V2); X1 = X2; Y1 != Y2; V1 = V2.
            """,
            """#program base.
subgrid_size(3).
size(S): S = (N*N) :- subgrid_size(N).
sudoku(X,Y,V) :- initial(X,Y,V).
1 = { sudoku(X,Y,V): V = (1..N) } :- size(N); X = (1..N); Y = (1..N).
__dom_sudoku(X,V) :- initial(X,_,V).
__dom_size((N*N)) :- subgrid_size(N).
__dom_sudoku(X,V) :- V = (1..N); __dom_size(N); X = (1..N); _ = (1..N).
#false :- __dom_sudoku(X2,V2); 2 <= #count { Y1: sudoku(X2,Y1,V2) }.""",
        ),
        (
            """
{max(P, X)} :- X = #max {V, ID : P=42, skill(P, ID, V); 23 : #true}, person(P), random(Y).
            """,  # currently not handled but in future, see #9
            """#program base.
{ max(P,X) } :- X = #max { V,ID: P = 42, skill(P,ID,V); 23 }; person(P); random(_).""",
        ),
        (
            """
grid(X, Y) :- X = 1..n, Y = 1..n.

{ ghost(X, Y) : grid(X, Y) } = N :- ghost(N).
{ dracula(X, Y) : grid(X, Y) } = N :- dracula(N).
{ zombie(X, Y) : grid(X, Y) } = N :- zombie(N).

:- ghost(X, Y), mirror(X, Y, T).
:- dracula(X, Y), mirror(X, Y, T).
:- zombie(X, Y), mirror(X, Y, T).
:- ghost(X, Y), dracula(X, Y).
:- ghost(X, Y), zombie(X, Y).
:- dracula(X, Y), zombie(X, Y).

ghost(X, Y, N) :- count(X, Y, TN),
         N = #count { GX, GY, D : mirror_visible(X, Y, GX, GY, D), ghost(GX, GY) }.
dracula(X, Y, N) :- count(X, Y, TN),
         N = #count { DX, DY, D : direct_visible(X, Y, DX, DY, D), dracula(DX, DY) }.
zombie(X, Y, N1 + N2) :- count(X, Y, TN),
         N1 = #count { ZX, ZY, D : mirror_visible(X, Y, ZX, ZY, D), zombie(ZX, ZY) },
         N2 = #count { ZX, ZY, D : direct_visible(X, Y, ZX, ZY, D), zombie(ZX, ZY) }.

:- count(X, Y, N), ghost(X, Y, G), dracula(X, Y, D), zombie(X, Y, Z), N != G + D + Z.

#show ghost/2.
#show dracula/2.
#show zombie/2.
            """,
            """#program base.
grid(X,Y) :- X = (1..n); Y = (1..n).
N = { ghost(X,Y): grid(X,Y) } :- ghost(N).
N = { dracula(X,Y): grid(X,Y) } :- dracula(N).
N = { zombie(X,Y): grid(X,Y) } :- zombie(N).
#false :- ghost(X,Y); mirror(X,Y,_).
#false :- dracula(X,Y); mirror(X,Y,_).
#false :- zombie(X,Y); mirror(X,Y,_).
#false :- ghost(X,Y); dracula(X,Y).
#false :- ghost(X,Y); zombie(X,Y).
#false :- dracula(X,Y); zombie(X,Y).
#false :- count(X,Y,N); 0 != #sum { (1*-1),ZX,ZY,D2,__agg(0),__agg(0): mirror_visible(X,Y,ZX,ZY,D2), zombie(ZX,ZY);\
 (1*-1),ZX,ZY,D2,__agg(1),__agg(0): direct_visible(X,Y,ZX,ZY,D2), zombie(ZX,ZY);\
 (1*-1),GX,GY,D0,__agg(1): mirror_visible(X,Y,GX,GY,D0), ghost(GX,GY);\
 (1*-1),DX,DY,D1,__agg(2): direct_visible(X,Y,DX,DY,D1), dracula(DX,DY); N,__agg(3) }.
#show ghost/2.
#show dracula/2.
#show zombie/2.""",
        ),
    ),
)
def test_all(lhs: str, rhs: str) -> None:
    """test removal of superseeded literals on whole programs"""
    prg: list[AST] = []
    parse_string(lhs, prg.append)
    prg = preprocess(prg)
    input_predicates = auto_detect_input(prg)

    prg = optimize(prg, input_predicates, [], duplication=True, projection=True)

    assert rhs == "\n".join(map(str, prg))



""" ngo lets you optimize your unground ASP encoding """

from copy import deepcopy
from typing import Iterable

from clingo.ast import AST

from ngo.cleanup import CleanupTranslator
from ngo.inline import InlineTranslator
from ngo.literal_duplication import LiteralDuplicationTranslator
from ngo.math_simplification import MathSimplification
from ngo.minmax_aggregates import MinMaxAggregator
from ngo.normalize import exline_arithmetic, postprocess, preprocess
from ngo.projection import ProjectionTranslator
from ngo.sum_aggregates import SumAggregator
from ngo.symmetry import SymmetryTranslator
from ngo.unused import UnusedTranslator
from ngo.utils.ast import Predicate


# pylint: disable=too-many-arguments
def optimize(
    prg: Iterable[AST],
    input_predicates: list[Predicate],
    output_predicates: list[Predicate],
    cleanup: bool = True,
    unused: bool = True,
    duplication: bool = False,
    symmetry: bool = True,
    minmax_chains: bool = True,
    sum_chains: bool = True,
    math: bool = True,
    inline: bool = True,
    projection: bool = True,
) -> list[AST]:
    """convert a logic program in form of clingo.AST's into an optimized encoding

    Parameters
    ----------
    prg: an encoding in form of clingo.AST's

    input_predicates : a list of `Predicate`'s that are supposed to be given from outside e.g.
    the instance for the encoding
    For example `[Predicate("path",2), Predicate("node",1)]`
    You can also let ngo try to automatically detect your
    input predicates with `ngo.utils.globals.auto_detect_input`

    output_predicates : a list of `Predicate`'s that are supposed to be outputted by the encoding.
    In general it is advised to reduce these predicates in your encoding to a minimum to optimize for performance.
    For example `[Predicate("selected",2), Predicate("chosen",1)]`
    You can also let ngo try to automatically detect your
    output predicates with -> `ngo.utils.globals.auto_detect_output`

    Traits:
    There is a selection of different traits available that can be turned on or off.
    Please see [Traits](#traits) for a detailed description.
    """
    input_: list[AST] = preprocess(prg)
    while True:
        old = deepcopy(input_)
        ### call transformers
        if cleanup:
            clt = CleanupTranslator(input_predicates)
            input_ = clt.execute(input_)

        if unused:
            utr = UnusedTranslator(input_, input_predicates, output_predicates)
            input_ = utr.execute(input_)

        if duplication:
            ldt = LiteralDuplicationTranslator(input_, input_predicates)
            input_ = ldt.execute(input_)

        if symmetry:
            trans = SymmetryTranslator(input_, input_predicates)
            input_ = trans.execute(input_)

        if minmax_chains:
            mma = MinMaxAggregator(input_, input_predicates)
            input_ = mma.execute(input_)

        if sum_chains:
            sagg = SumAggregator(input_, input_predicates)
            input_ = sagg.execute(input_)

        if math:
            mmath = MathSimplification(input_)
            input_ = mmath.execute(input_)

        if inline:
            inl = InlineTranslator(input_, input_predicates, output_predicates)
            input_ = inl.execute(input_)

        if projection:
            pro = ProjectionTranslator(input_, input_predicates)
            input_ = pro.execute(input_)

        input_ = exline_arithmetic(input_)

        if input_ == old:
            break
    input_ = postprocess(input_)
    return input_



"""
Test cases for main application functionality.
"""
from argparse import ArgumentTypeError

import pytest

from ngo.utils.ast import Predicate
from ngo.utils.parser import ALL_OPTIONS, DEFAULT_OPTIONS, get_parser


def test_parser() -> None:
    """
    Test the parser.
    """
    parser = get_parser()
    args = parser.parse_args(["--log", "error"])
    args = parser.parse_args(["--enable", "minmax_chains", "math"])
    assert "math" in args.enable
    assert "minmax_chains" in args.enable
    args = parser.parse_args(["--enable", "all"])
    assert ALL_OPTIONS == args.enable
    args = parser.parse_args(["--enable", "default"])
    assert sorted(DEFAULT_OPTIONS) == sorted(args.enable)

    with pytest.raises(ArgumentTypeError) as excinfo:
        parser.parse_args(["--enable", "none", "math"])
    assert str(excinfo.value) == "'none' may not be combined with other options."
    args = parser.parse_args(["--input-predicates", "zero/0, another/14"])
    assert Predicate("zero", 0) in args.input_predicates
    assert Predicate("another", 14) in args.input_predicates
    args = parser.parse_args(["--input-predicates", "auto"])
    assert args.input_predicates == "auto"
    args = parser.parse_args(["--input-predicates", ""])
    assert len(args.input_predicates) == 0
    args = parser.parse_args(["--input-predicates"])
    assert len(args.input_predicates) == 0
    with pytest.raises(ArgumentTypeError) as excinfo:
        parser.parse_args(["--input-predicates", "zero"])
    assert str(excinfo.value) == "Predicates have to be of the form name/arity, but got zero."
    with pytest.raises(ArgumentTypeError) as excinfo:
        parser.parse_args(["--input-predicates", "zero/one"])
    assert str(excinfo.value) == "Arity of predicate zero should be integer, but got one."



"""
 This module inlines predicates using aggregates that are only used in one rule
"""

import logging
from itertools import permutations
from typing import Iterable, Optional

import networkx as nx
from clingo import Number
from clingo.ast import AST, AggregateFunction, ASTType, Function, Sign, SymbolicTerm, Variable

from ngo.dependency import DomainPredicates, RuleDependency
from ngo.utils.ast import (
    LOC,
    SIGNS,
    AggAnalytics,
    Predicate,
    collect_ast,
    global_vars_inside_body,
    global_vars_inside_head,
    is_predicate,
    literal_predicate,
    potentially_unifying_sequence,
    transform_ast,
)
from ngo.utils.globals import UniqueNames, UniqueVariables

log = logging.getLogger(__name__)


class InlineTranslator:
    """Inlines rules that are the q unique reason for a predicate"""

    def __init__(self, prg: list[AST], input_predicates: list[Predicate], output_predicates: list[Predicate]):
        self.unique_names = UniqueNames(prg, input_predicates)
        self.domain_predicates = DomainPredicates(self.unique_names, prg)
        self.input_predicates = input_predicates
        self.output_predicates = output_predicates
        self.minimize_tuples: list[list[AST]] = []

    def analyze_minimize(self, prg: list[AST]) -> None:
        """analyze progra, and collect all minimize tuples"""
        self.minimize_tuples = []
        for stm in prg:
            if stm.ast_type == ASTType.Minimize:
                self.minimize_tuples.append([stm.weight, stm.priority] + list(stm.terms))

    @staticmethod
    def _info(rule: AST) -> tuple[int, int, int]:
        """return number of aggregates, number of conditional literals and number of literals in body"""

        def _num_aggs(blit: AST) -> int:
            return len(collect_ast(blit, "BodyAggregate")) + len(collect_ast(blit, "Aggregate"))

        def _num_conditionals(blit: AST) -> bool:
            return len(collect_ast(blit, "ConditionalLiteral")) > 0

        num_aggs = sum(map(_num_aggs, rule.body))
        num_conditionals = any(map(_num_conditionals, rule.body))
        num_lits = len(rule.body)

        return num_aggs, num_conditionals, num_lits

    @staticmethod
    def transform_args(orig: list[AST], passed: list[AST], asts: list[AST], unique_vars: UniqueVariables) -> list[AST]:
        """given a list of variables of the original predicate
        and a list of arguments passed to the predicate
        replace all the variables inside asts with the passed arguments"""
        orig2passed: dict[AST, AST] = dict(zip(orig, passed))

        def trans(var: AST) -> AST:
            if var not in orig2passed:
                orig2passed[var] = unique_vars.make_unique(var)
            return orig2passed[var]

        new_asts: list[AST] = []
        for ast in asts:
            new_asts.append(transform_ast(ast, "Variable", trans))
        return new_asts

    def inline_literal(self, rule: AST, lit: AST, unique_vars: UniqueVariables) -> list[AST]:
        """inline rule into this body literal"""
        orig_arguments = list(rule.head.atom.symbol.arguments)
        assert lit.ast_type == ASTType.Literal
        atom = lit.atom
        assert atom.ast_type == ASTType.SymbolicAtom and atom.symbol.ast_type == ASTType.Function

        passed_arguments = list(atom.symbol.arguments)
        if lit.sign == Sign.Negation:
            newlit = rule.body[0]
            newlit = newlit.update(sign=Sign.Negation)
            newlit = newlit.update(
                atom=self.transform_args(orig_arguments, passed_arguments, [newlit.atom], unique_vars)[0]
            )
            return [newlit]
        # NoSign
        return self.transform_args(orig_arguments, passed_arguments, rule.body, unique_vars=unique_vars)

    def compute_new_body_elements(
        self, rule: AST, replace_cond: AST, replace_elem: AST, agg: AST, atom: AST, unique_vars: UniqueVariables
    ) -> list[AST]:
        """create new body elements from the aggregate, replacing the weight of the aggregate"""
        hatom = rule.head.atom
        max_arity: int = 0
        for elem in atom.elements:
            max_arity = max(max_arity, len(elem.condition))
        rbody = [
            blit
            for blit in rule.body
            if not (blit.ast_type == ASTType.Literal and blit.atom.ast_type == ASTType.BodyAggregate)
        ]
        new_elements: list[AST] = []
        orig = hatom.symbol.arguments
        pass_ = replace_cond.atom.symbol.arguments
        for elem in agg.elements:
            transformed = self.transform_args(  # transform all lists at the same time to get equal aux vars
                orig,
                pass_,
                list(elem.terms) + rbody + list(elem.condition),
                unique_vars,
            )
            terms = transformed[: len(elem.terms)]
            transformed = transformed[len(elem.terms) :]
            new_terms = terms + list(replace_elem.terms[1:])
            new_terms.extend([Function(LOC, "unique", [], False)] * (max_arity - len(new_terms) + 1))
            new_elements.append(elem.update(terms=new_terms, condition=transformed))
        return new_elements

    def inline_body_aggregate(self, rule: AST, atom: AST, unique_vars: UniqueVariables) -> AST:
        """inline rule into this body aggregate atom"""
        # pylint: disable=too-many-branches
        hatom = rule.head.atom
        hpred = Predicate(hatom.symbol.name, len(hatom.symbol.arguments))
        num_aggs, num_cond, _ = self._info(rule)

        if num_cond == 0 and num_aggs == 1:
            # result of aggregate in rule may only be used in the head variable, called HV
            agg = collect_ast(rule, "BodyAggregate")[0]

            good = {
                AggregateFunction.Min: (AggregateFunction.Min,),
                AggregateFunction.Max: (AggregateFunction.Max,),
                AggregateFunction.Count: (
                    AggregateFunction.Count,
                    AggregateFunction.Sum,
                    AggregateFunction.SumPlus,
                ),
                AggregateFunction.Sum: (
                    AggregateFunction.Sum,
                    AggregateFunction.SumPlus,
                ),
                AggregateFunction.SumPlus: (
                    AggregateFunction.Sum,
                    AggregateFunction.SumPlus,
                ),
            }
            result_function = atom.function
            if agg.function == AggregateFunction.Sum:
                result_function = AggregateFunction.Sum
            if atom.function not in good[agg.function]:
                return atom
            agga = AggAnalytics(agg)
            # result is actually used in head
            for hv_pos, hv in enumerate(hatom.symbol.arguments):
                if hv == Variable(LOC, agga.equal_variable_bound[0]):
                    break
            else:
                assert True, "sinle_rule should already check this"
            # result is not used anywhere else (1 time in the head, 1 time in aggregate equal)
            if sum(map(lambda x: x == hv, collect_ast(rule, "Variable"))) != 2:
                return atom

            replace_elem: Optional[AST] = None
            replace_cond: AST
            for elem in atom.elements:
                for cond in elem.condition:
                    if hpred in set(p.pred for p in literal_predicate(cond, SIGNS)):
                        replace_elem = elem
                        replace_cond = cond
            if replace_elem is None or replace_cond.ast_type != ASTType.Literal or replace_cond.sign != Sign.NoSign:
                return atom
            rest_elems = [elem for elem in atom.elements if elem != replace_elem]
            ### check if tuple set semantic does not allow for unique identification
            if any(map(lambda x: potentially_unifying_sequence(x.terms, replace_elem.terms), rest_elems)):
                log.info(f"Cannot inline {str(hpred)} into {str(atom)} as the tuple is not unique.")
                return atom
            # check if exactly the HV variable is also used as the weight in the new element
            if (
                not replace_elem.terms
                or replace_elem.terms[0]
                != replace_cond.atom.symbol.arguments[hv_pos]  # pylint: disable=undefined-loop-variable
            ):
                return atom
            # replace headrule body aggregate with inlined version of the conditions
            new_elements = self.compute_new_body_elements(rule, replace_cond, replace_elem, agg, atom, unique_vars)
            return atom.update(function=result_function, elements=rest_elems + new_elements)
        return atom

    def inline_minimize(self, stm: AST) -> list[AST]:
        """inline sum/count that are inside weak constraint
        return possible more weak constraints"""
        if stm.ast_type != ASTType.Minimize:
            return [stm]
        # find first aggregate and equality variable
        # result of aggregate is called HV
        aggs = collect_ast(stm, "BodyAggregate")
        if not aggs:
            return [stm]
        agg = aggs[0]
        # TODO: i do need to rename local variables as they become global now, if several aggregates are used
        if len(aggs) > 1:
            return [stm]

        if agg.function not in (AggregateFunction.Count, AggregateFunction.Sum, AggregateFunction.SumPlus):
            return [stm]
        agga = AggAnalytics(agg)
        # only one equality
        if len(agga.equal_variable_bound) != 1 or agga.bounds:
            return [stm]
        # result is actually used as weight
        hv = Variable(LOC, agga.equal_variable_bound[0])
        if hv != stm.weight:
            return [stm]

        # result is not used anywhere else (1 time in the head, 1 time in aggregate equal)
        if sum(map(lambda x: x == hv, collect_ast(stm, "Variable"))) != 2:
            return [stm]

        ### check if tuple set semantic does not allow for unique identification
        replace_terms = [stm.weight, stm.priority] + list(stm.terms)
        if any(
            map(
                lambda x: potentially_unifying_sequence(x, replace_terms),
                [t for t in self.minimize_tuples if t != replace_terms],
            )
        ):
            log.info(f"Cannot inline agregate into {str(stm)} as the tuple is not unique.")
            return [stm]

        # replace body aggregate with inlined version of the conditions
        rbody = [blit for blit in stm.body if not (blit.ast_type == ASTType.Literal and blit.atom == agg)]
        new_minimizes = []
        max_arity = 0
        for tuple_ in self.minimize_tuples:
            max_arity = max(max_arity, len(tuple_))
        max_arity -= 2  # for weight and priority
        for elem in agg.elements:
            # replace old weight with tuple from aggregate
            new_weight: AST
            if agg.function == AggregateFunction.Count:
                new_weight = SymbolicTerm(LOC, Number(1))  # nocoverage
            else:
                new_weight = elem.terms[0]
            # create new minimize with this elem replaced by its content
            new_terms = list(elem.terms[1:]) + list(stm.terms)
            new_terms.extend([Function(LOC, "unique", [], False)] * (max_arity - len(new_terms) + 1))
            new_body = rbody + list(elem.condition)
            new_minimizes.append(stm.update(body=new_body, terms=new_terms, weight=new_weight))
        return new_minimizes

    @staticmethod
    def has_anonymous_vars(pred: Predicate, body: list[AST]) -> bool:
        """return true if pred inside stm uses anonymous variables"""
        for lit in body:
            if is_predicate(lit) and Predicate(lit.atom.symbol.name, len(lit.atom.symbol.arguments)) == pred:
                if any(map(lambda x: x == Variable(LOC, "_"), lit.atom.symbol.arguments)):
                    return True
        return False

    def is_single(self, stm: AST, rdp: RuleDependency) -> Optional[int]:
        """if the stm is the only statement in the program producing this predicate
        and it also contains the result of an aggregate, return the position of this result,
        otherwise None"""
        if stm.ast_type != ASTType.Rule:
            return None
        # check if head is simple predicate
        ## TODO: refactor to check if literal is predicate
        if not is_predicate(stm.head) or stm.head.ast_type == ASTType.Literal and stm.head.sign != Sign.NoSign:
            return None
        hatom = stm.head.atom
        hpred = Predicate(hatom.symbol.name, len(hatom.symbol.arguments))

        # no duplicates as this would be an implicit constraint
        if any(map(lambda arg: arg.ast_type != ASTType.Variable, hatom.symbol.arguments)) or len(
            set(collect_ast(hatom.symbol, "Variable"))
        ) != len(hatom.symbol.arguments):
            return None

        if hpred in self.input_predicates or hpred in self.output_predicates or self.domain_predicates.is_static(hpred):
            return None
        deriving_rules = rdp.get_rules_that_derive(hpred)
        if len(deriving_rules) != 1:
            return None

        using_stms = rdp.get_statements_that_use(hpred)
        if len(using_stms) != 1 or using_stms[0] == stm or self.has_anonymous_vars(hpred, using_stms[0].body):
            return None

        # result of aggregate must be inside head
        aggs = collect_ast(stm, "BodyAggregate")
        if not aggs or len(aggs) != 1:
            return None
        agg = aggs[0]

        agga = AggAnalytics(agg)
        # only one equality
        if len(agga.equal_variable_bound) != 1 or agga.bounds:
            return None
        # result is actually used in head
        for index, v in enumerate(hatom.symbol.arguments):
            if v == Variable(LOC, agga.equal_variable_bound[0]):
                return index
        return None

    def get_body_lit(self, stm: AST, orig: AST) -> Optional[AST]:
        """given the single stm and a rule where it occurs in, return the body predicate if available"""
        hatom = stm.head.atom
        hpred = Predicate(hatom.symbol.name, len(hatom.symbol.arguments))

        blit: AST
        for blit in orig.body:
            if not is_predicate(blit):
                continue
            atom = blit.atom

            if hpred != Predicate(atom.symbol.name, len(atom.symbol.arguments)):
                continue
            # passed_arguments = list(atom.symbol.arguments)
            if blit.sign == Sign.DoubleNegation:
                continue  # nocoverage
            if blit.sign == Sign.Negation:
                # should not introduce new variables
                if len(stm.body) > 1 or global_vars_inside_head(stm.head) != global_vars_inside_body(stm.body):
                    continue
                # i need to extra negate this literal, not sure what happens with not not ...
                if stm.body[0].sign != Sign.NoSign:
                    continue  # nocoverage
                return blit
            return blit
        return None

    def is_connected_to_agregates(self, var: AST, orig: AST, g: nx.Graph) -> bool:
        """True if var is connected to an aggregate via arithmetic"""
        vargraph = nx.Graph()

        def add(varlist: Iterable[AST]) -> None:
            """add all vars to g as a node
            increment "aggr" attribute by one"""
            for var in varlist:
                vargraph.add_node(var)
                if "aggr" in vargraph.nodes[var]:
                    vargraph.nodes[var]["aggr"] += 1
                else:
                    vargraph.nodes[var]["aggr"] = 1

        globals_ = global_vars_inside_body(orig.body)
        for blit in orig.body:
            if (orig, blit) in g.nodes:
                # single = g.nodes[(orig, blit)]["stm"]
                add([blit.atom.symbol.arguments[g.nodes[(orig, blit)]["index"]]])
            elif blit.ast_type == ASTType.Literal:
                atom = blit.atom
                if atom.ast_type == ASTType.BodyAggregate:
                    if atom.left_guard:
                        add(collect_ast(atom.left_guard, "Variable"))
                    if atom.right_guard:
                        add(collect_ast(atom.right_guard, "Variable"))
                    add(globals_.intersection(set(collect_ast(atom, "Variable"))))
                elif atom.ast_type == ASTType.Comparison:
                    vargraph.add_edges_from(permutations(collect_ast(atom, "Variable"), 2))
        if orig.ast_type == ASTType.Minimize:
            add(collect_ast(orig.weight, "Variable"))
        for n in vargraph.nodes:
            vargraph.add_edge(n, n)
        for cc in nx.connected_components(vargraph):
            if var not in cc:
                continue
            if sum(map(lambda n: vargraph.nodes[n]["aggr"] if "aggr" in vargraph.nodes[n] else 0, cc)) > 1:
                return True
        return False

    def replace_single_rule_for_body(self, prg: list[AST]) -> list[AST]:
        """replace a single rule that has a predicate that only they produce
        and that is only used once somewhere else in the program
        Also the predicate must contain the result of an aggregate computation
        Also the predicate must be used inside an body literal and its arithmetic is connected to another aggregate
        Does not replace any rule if not possible"""
        g = nx.Graph()
        rdp = RuleDependency(prg)
        for stm in prg:
            index = self.is_single(stm, rdp)
            if index is None:
                continue

            hatom = stm.head.atom
            hpred = Predicate(hatom.symbol.name, len(hatom.symbol.arguments))
            orig = rdp.get_statements_that_use(hpred)[0]

            blit = self.get_body_lit(stm, orig)
            if blit is None:
                continue
            g.add_node((orig, blit), stm=stm, index=index)
        na_stm = nx.get_node_attributes(g, "stm")
        na_index = nx.get_node_attributes(g, "index")
        for n in g.nodes:
            orig, blit_ = n
            stm = na_stm[n]  # can't access attributes in the normal way as tuple [] operator blocks access
            index_ = na_index[n]

            if not self.is_connected_to_agregates(blit_.atom.symbol.arguments[index_], orig, g):
                continue
            replace = self.inline_literal(stm, blit_, UniqueVariables(orig))
            if [blit_] != replace:
                new_prg: list[AST] = []
                for r in prg:
                    if r == stm:  # skip singular rule
                        continue
                    if r != orig:
                        new_prg.append(r)
                        continue
                    # replace orig
                    new_body: list[AST] = []
                    for lit in orig.body:
                        if lit == blit_:
                            new_body.extend(replace)
                        else:
                            new_body.append(lit)
                    new_prg.append(orig.update(body=new_body))
                return new_prg
        return prg

    def replace_inside_agg(self, stm: AST, orig: AST) -> AST:
        """given a single rule and an index to the aggregates position,
        return the original rule and its replacement if the singular rule replaces something in an aggregate"""
        unique_vars = UniqueVariables(orig)
        if orig.ast_type != ASTType.Rule:
            return orig
        new_body: list[AST] = []
        for blit in orig.body:
            if blit.ast_type == ASTType.Literal and blit.atom.ast_type == ASTType.BodyAggregate:
                new_body.append(blit.update(atom=self.inline_body_aggregate(stm, blit.atom, unique_vars)))
            else:
                new_body.append(blit)
        return orig.update(body=new_body)

    def replace_single_rule_for_agg(self, prg: list[AST]) -> list[AST]:
        """replace a single rule that has a predicate that only they produce
        and that is only used once somewhere else in the program
        Also the predicate must contain the result of an aggregate computation
        Also the predicate must be used inside an aggregate
        Does not replace any rule if not possible"""
        rdp = RuleDependency(prg)
        for stm in prg:
            index = self.is_single(stm, rdp)
            if index is None:
                continue
            hatom = stm.head.atom
            hpred = Predicate(hatom.symbol.name, len(hatom.symbol.arguments))
            orig = rdp.get_statements_that_use(hpred)[0]

            replace = self.replace_inside_agg(stm, orig)
            if orig != replace:
                new_prg: list[AST] = []
                for r in prg:
                    if r == stm:  # skip singular rule
                        continue
                    if r == orig:  # replace orig
                        new_prg.append(replace)
                        continue
                    new_prg.append(r)
                return new_prg
        return prg

    def inline_in_agg(self, prg: list[AST]) -> list[AST]:
        """inline single rules that contain an aggregate into other aggregates"""
        while True:
            new_prg = self.replace_single_rule_for_agg(prg)
            if new_prg == prg:
                break
            prg = new_prg
        return prg

    def inline_in_rulebody(self, prg: list[AST]) -> list[AST]:
        """inline single rules that contain an aggregate into a rule body or objective
        if the aggregates value is used in a computation involving an aggregate or objective"""
        while True:
            new_prg = self.replace_single_rule_for_body(prg)
            if new_prg == prg:
                break
            prg = new_prg
        return prg

    def inline_in_minimize(self, prg: list[AST]) -> list[AST]:
        """inline single aggregates inside an objective into its weight"""
        self.analyze_minimize(prg)

        new_prg = []
        for stm in prg:
            new_prg.extend(self.inline_minimize(stm))
        return new_prg

    def execute(self, prg: list[AST]) -> list[AST]:
        """
        remove all rules that are unique and can be inlined
        """
        prg = self.inline_in_agg(prg)
        prg = self.inline_in_rulebody(prg)
        prg = self.inline_in_minimize(prg)
        return prg



"""
 This module projects out variables from rules to reduce grounding complexity
"""

import logging
from typing import Optional

from clingo.ast import AST, ASTType, Function, Literal, Rule, Sign, SymbolicAtom, Variable

from ngo.normalize import inline_arithmetic
from ngo.utils.ast import (
    LOC,
    Predicate,
    collect_ast,
    collect_binding_information_body,
    global_vars_inside_body,
    global_vars_inside_head,
    is_predicate,
    largest_subset,
)
from ngo.utils.globals import UniqueNames

log = logging.getLogger(__name__)


class ProjectionTranslator:
    """Splits rules to reduce grounding complexity"""

    def __init__(self, prg: list[AST], input_predicates: list[Predicate]):
        self.unique_names = UniqueNames(prg, input_predicates)

    def good_split(self, new: list[AST], rest: list[AST], stm: AST) -> Optional[list[AST]]:
        """given a new subset of the body of a rule
        determine if this is a good split and if yes, return
        the necessary variables"""
        ## some split that is legal
        if not (1 < len(new) < len(stm.body)) and len(rest):
            return None
        # new rule is legal
        _, unbound = collect_binding_information_body(new)
        if unbound:
            return None

        # staying rule still legal
        vars_in_rest: set[AST] = set()
        for r in rest:
            vars_in_rest.update(collect_ast(r, "Variable"))
        vars_in_rest.discard(Variable(LOC, "_"))
        t = global_vars_inside_body(new).intersection(vars_in_rest | global_vars_inside_head(stm.head))
        if collect_binding_information_body(rest, t)[1]:
            return None  # nocoverage

        ## global vars can not become unglobal
        # if variable in new is there but not global in new but has been global before, this is bad
        vars_in_new: set[AST] = set()
        for r in new:
            vars_in_new.update(collect_ast(r, "Variable"))
        vars_in_new.discard(Variable(LOC, "_"))
        local_new = vars_in_new.difference(global_vars_inside_body(new))
        global_old = global_vars_inside_body(stm.body)
        if local_new.intersection(global_old):
            return None
        # if variable in updated rule is there but not global but was global before this is bad
        # local_new = (t | vars_in_rest).difference()

        ## split is useful
        # changed and new rule safe a variable
        if len(t | vars_in_rest) >= len(global_old) or len(t) >= len(global_vars_inside_body(new)):
            return None

        # dont leave literals behind that do not introduce new variables
        for r in rest:
            if (set(collect_ast(r, "Variable")) - {Variable(LOC, "_")}).issubset(vars_in_new):
                return None

        # rest contains at least one symbolic literal that is true
        def aux(lit: AST) -> bool:
            """return true if true symbolic literal"""
            return is_predicate(lit) and lit.sign == Sign.NoSign

        if not any(map(aux, rest)):
            return None

        # dont split up aggregates
        new_aggs = any(map(lambda x: len(collect_ast(x, "BodyAggregate")) > 0, new))
        rest_aggs = any(map(lambda x: len(collect_ast(x, "BodyAggregate")) > 0, rest))
        if new_aggs and rest_aggs:
            return None

        # only split if new rule has less head variables than old rule
        if len(t) >= len(global_vars_inside_head(stm.head)):
            return None
        return sorted(t)

    def project_rule(self, stm: AST) -> list[AST]:
        """replace a rule if a splitted version if this improves grounding complexity"""
        assert stm.ast_type == ASTType.Rule
        for new_list in largest_subset(stm.body):
            new: list[AST] = list(new_list)
            rest = [x for x in stm.body if x not in new]
            split_vars = self.good_split(new, rest, stm)
            if split_vars is None:
                continue
            aux_pred = self.unique_names.new_auxpredicate(len(split_vars))
            logging.info(f"Split rule {stm} with new predicate {aux_pred}")
            new_rule = Rule(
                LOC, Literal(LOC, Sign.NoSign, SymbolicAtom(Function(LOC, aux_pred.name, split_vars, False))), new
            )
            updated_rule = stm.update(body=rest + [new_rule.head])
            return [new_rule, updated_rule]
        return [stm]

    def execute(self, prg: list[AST]) -> list[AST]:
        """
        remove all rules that are unique and can be inlined
        """
        ret: list[AST] = []
        prg = inline_arithmetic(prg)
        for stm in prg:
            if stm.ast_type == ASTType.Rule:
                ret.extend(self.project_rule(stm))
            else:
                ret.append(stm)
        return ret



""" general util functions and classes """

import logging
from collections import defaultdict
from itertools import chain
from typing import Iterable

from clingo.ast import AST, ASTType, Variable

from ngo.utils.ast import (
    LOC,
    SIGNS,
    Predicate,
    body_predicates,
    collect_ast,
    headderivable_predicates,
    minimize_predicates,
    predicates,
)

AUX_FUNC = "__aux_"
CHAIN_STR = "__chain"
MIN_STR = "__min_"
MAX_STR = "__max_"
NEXT_STR = "__next_"
DOM_STR = "__dom_"
AGG_STR = "__agg"
NEXT = Variable(LOC, "__NEXT")
PREV = Variable(LOC, "__PREV")
AUX_VAR = Variable(LOC, "AUX")

log = logging.getLogger(__name__)


def auto_detect_input(prg: Iterable[AST]) -> list[Predicate]:
    """
    given a program return a list of all predicates that occur in the program
    but are not derivable in a head
    """
    all_preds: set[Predicate] = set()
    derivable_preds: set[Predicate] = set()
    in_body: dict[Predicate, set[int]] = defaultdict(set)
    in_head: dict[Predicate, set[int]] = defaultdict(set)
    for index, stm in enumerate(prg):
        all_preds.update([pred.pred for pred in predicates(stm)])
        for pred in headderivable_predicates(stm):
            derivable_preds.add(pred.pred)
            in_head[pred.pred].add(index)
        for pred in chain(body_predicates(stm, SIGNS), minimize_predicates(stm, SIGNS)):
            in_body[pred.pred].add(index)

    input_ = list(sorted(all_preds - derivable_preds))
    for p in all_preds:
        if in_body[p] == in_head[p]:
            input_.append(p)
    for p in input_:
        log.info(f"Detected input predicate: {p.name}/{p.arity}")
    return input_


def auto_detect_output(prg: Iterable[AST]) -> list[Predicate]:
    """
    given a program return a list of all predicates used in show statements
    """
    output: set[Predicate] = set()
    for stm in prg:
        if stm.ast_type == ASTType.ShowSignature:
            output.add(Predicate(stm.name, stm.arity))
        elif stm.ast_type == ASTType.ShowTerm:
            for lit in stm.body:
                output.update([p.pred for p in predicates(lit)])
    if output:
        log.info(
            "Output detected. Consider using a postprocessor to format your output instead of rules and statements."
        )
    for pred in sorted(output):
        log.info(f"Detected output predicate: {pred.name}/{pred.arity}")
    return list(sorted(output))


class UniqueVariables:
    """class to provide unique names for variables in a rule/minimize"""

    def __init__(self, rule: AST) -> None:
        self._allvars: list[AST] = collect_ast(rule, "Variable")

    def make_unique(self, var: AST) -> AST:
        """return itself if not already present in rule, otherwise
        add a counter to it to make it unique
        also make it known so that it stays unique"""
        if var.name == "_":
            return var
        if var not in self._allvars:
            self._allvars.append(var)
            return var
        count = 0
        while True:
            newvar = var.update(name=var.name + str(count))
            if newvar not in self._allvars:
                self._allvars.append(newvar)
                return newvar
            count += 1


class UniqueNames:
    """class to provide unique names for predicates, functions"""

    def __init__(self, prg: list[AST], input_predicates: list[Predicate]) -> None:
        self.auxcounter = 0
        self.predicates: set[Predicate] = set(input_predicates)
        for stm in prg:
            for spred in predicates(stm):
                self.predicates.add(spred.pred)

    def new_auxpredicate(self, arity: int) -> Predicate:
        """provide a unique aux Predicate"""
        self.auxcounter += 1
        p = Predicate(AUX_FUNC + str(self.auxcounter), arity)
        while p in self.predicates:
            p = Predicate(AUX_FUNC + str(self.auxcounter), arity)
            self.auxcounter += 1
        self.predicates.add(p)
        return p

    def new_predicate(self, similar: str, arity: int) -> Predicate:
        """provide a Predicate that is similar but unique"""
        p = Predicate(similar, arity)
        counter = 1
        while p in self.predicates:
            p = Predicate(similar + str(counter), arity)
            counter += 1
        self.predicates.add(p)
        return p
